from arcgis.gis import Layer, _GISResource


class NetworkLayer(Layer):
    """
    NetworkLayer represents a single network layer. It provides basic
    information about the network layer such as its name, type, and network
    classes. Additionally, depending on the layer type, it provides different
    pieces of information.

    It is a base class for RouteLayer, ServiceAreaLayer, and
    ClosestFacilityLayer.
    """
    def retrieve_travel_modes(self):
        """identify all the valid travel modes that have been defined on the
        network dataset or in the portal if the GIS server is federated"""
        url = self._url + "/retrieveTravelModes"
        params = {"f":"json"}
        return self._con.get(path=url,
                         params=params, token=self._token)


class RouteLayer(NetworkLayer):
    """
    The Route Layer which has common properties of Network Layer
    as well as some attributes unique to Route Network Layer only.
    """
    def solve(self,
              stops,
              barriers=None,
              polyline_barriers=None,
              polygon_barriers=None,
              travel_mode=None,
              attribute_parameter_values=None,
              return_directions=None,
              return_routes=True,
              return_stops=False,
              return_barriers=False,
              return_polyline_barriers=True,
              return_polygon_barriers=True,
              out_sr=None,
              ignore_invalid_locations=True,
              output_lines=None,
              find_best_sequence=False,
              preserve_first_stop=True,
              preserve_last_stop=True,
              use_time_windows=False,
              start_time=None,
              start_time_is_utc=False,
              accumulate_attribute_names=None,
              impedance_attribute_name=None,
              restriction_attribute_names=None,
              restrict_u_turns=None,
              use_hierarchy=True,
              directions_language=None,
              directions_output_type=None,
              directions_style_name=None,
              directions_length_units=None,
              directions_time_attribute_name=None,
              output_geometry_precision=None,
              output_geometry_precision_units=None,
              return_z=False
              ):
        """
        The solve operation is performed on a network layer resource.
        The solve operation is supported on a network layer whose layerType
        is esriNAServerRouteLayer. You can provide arguments to the solve
        route operation as query parameters.


        ===================================     ====================================================================
        **Argument**                            **Description**
        -----------------------------------     --------------------------------------------------------------------
        stops                                   Required Points/FeatureSet. The set of stops loaded as network
                                                locations during analysis. Stops can be specified using a simple
                                                comma / semi-colon based syntax or as a JSON structure. If stops are
                                                not specified, preloaded stops from the map document are used in the
                                                analysis.
        -----------------------------------     --------------------------------------------------------------------
        barriers                                Optional Point/FeatureSet. The set of barriers loaded as network
                                                locations during analysis. Barriers can be specified using a simple
                                                comma/semi-colon based syntax or as a JSON structure. If barriers
                                                are not specified, preloaded barriers from the map document are used
                                                in the analysis. If an empty json object is passed ('{}') preloaded
                                                barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        polyline_barriers                       Optional Polyline/FeatureSet. The set of polyline barriers loaded
                                                as network locations during analysis. If polyline barriers are not
                                                specified, preloaded polyline barriers from the map document are
                                                used in the analysis. If an empty json object is passed ('{}')
                                                preloaded polyline barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        polygon_barriers                        Optional Polygon/FeatureSet. The set of polygon barriers loaded as
                                                network locations during analysis. If polygon barriers are not
                                                specified, preloaded polygon barriers from the map document are used
                                                in the analysis. If an empty json object is passed ('{}') preloaded
                                                polygon barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        travel_mode                             Optional string. Travel modes provide override values that help you
                                                quickly and consistently model a vehicle or mode of transportation.
                                                The chosen travel mode must be preconfigured on the network dataset
                                                that the routing service references.
        -----------------------------------     --------------------------------------------------------------------
        attribute_parameter_values              Optional string/list.  A set of attribute parameter values that can be
                                                parameterized to determine which network elements can be used by a
                                                vehicle.
        -----------------------------------     --------------------------------------------------------------------
        return_directions                       Optional boolean. If true, directions will be generated and returned
                                                with the analysis results. Default is true.
        -----------------------------------     --------------------------------------------------------------------
        return_routes                           Optional boolean. If true, routes will be returned with the analysis
                                                results. Default is true.
        -----------------------------------     --------------------------------------------------------------------
        return_stops                            Optional boolean.  If true, stops will be returned with the analysis
                                                results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_barriers                         Optional boolean.  If true, barriers will be returned with the
                                                analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_polyline_barriers                Optional boolean. If true, polyline barriers will be returned with
                                                the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_polygon_barriers                 Optional boolean. If true, polygon barriers will be returned with
                                                the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        out_sr                                  Optional Integer. The spatial reference of the geometries returned
                                                with the analysis results.
        -----------------------------------     --------------------------------------------------------------------
        ignore_invalid_locations                Optional boolean. - If true, the solver will ignore invalid
                                                locations. Otherwise, it will raise an error. The default is as
                                                defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        output_lines                            The type of output lines to be generated in the result. The default
                                                is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        find_best_sequence                      Optional boolean. If true, the solver should re-sequence the route in
                                                the optimal order. The default is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        preserve_first_stop                     Optional boolean. If true, the solver should keep the first stop
                                                fixed in the sequence. The default is as defined in the network
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        preserve_last_stop                      Optional boolean. If true, the solver should keep the last stop fixed
                                                in the sequence. The default is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        use_time_window                         Optional boolean. If true, the solver should consider time windows.
                                                The default is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        start_time                              Optional string. The time the route begins. If not specified, the
                                                solver will use the default as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        start_time_is_utc                       Optional boolean. The time zone of the startTime parameter.
        -----------------------------------     --------------------------------------------------------------------
        accumulate_attribute_names              Optional string. A list of network attribute names to be accumulated
                                                with the analysis. The default is as defined in the network layer.
                                                The value should be specified as a comma separated list of attribute
                                                names. You can also specify a value of none to indicate that no
                                                network attributes should be accumulated.
        -----------------------------------     --------------------------------------------------------------------
        impedance_attribute_name                Optional string. The network attribute name to be used as the impedance
                                                attribute in analysis. The default is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        restriction_attribute_names             Optional string. -The list of network attribute names to be
                                                used as restrictions with the analysis. The default is as defined in
                                                the network layer. The value should be specified as a comma
                                                separated list of attribute names. You can also specify a value of
                                                none to indicate that no network attributes should be used as
                                                restrictions.
        -----------------------------------     --------------------------------------------------------------------
        restrict_u_turns                        Optional boolean. Specifies how U-Turns should be restricted in the
                                                analysis. The default is as defined in the network layer.
                                                Values: esriNFSBAllowBacktrack | esriNFSBAtDeadEndsOnly |
                                                        esriNFSBNoBacktrack | esriNFSBAtDeadEndsAndIntersections
        -----------------------------------     --------------------------------------------------------------------
        use_hierarchy                           Optional boolean.  If true, the hierarchy attribute for the network
                                                should be used in analysis. The default is as defined in the network
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        directions_language                     Optional string. The language to be used when computing directions.
                                                The default is as defined in the network layer. The list of supported
                                                languages can be found in REST layer description.
        -----------------------------------     --------------------------------------------------------------------
        directions_output_type                  Optional string.  Defines content, verbosity of returned directions.
                                                The default is esriDOTStandard.
                                                Values: esriDOTComplete | esriDOTCompleteNoEvents
                                                        | esriDOTInstructionsOnly | esriDOTStandard |
                                                        esriDOTSummaryOnly
        -----------------------------------     --------------------------------------------------------------------
        directions_style_name                   Optional string. The style to be used when returning the directions.
                                                The default is as defined in the network layer. The list of
                                                supported styles can be found in REST layer description.
        -----------------------------------     --------------------------------------------------------------------
        directions_length_units                 Optional string. The length units to use when computing directions.
                                                The default is as defined in the network layer.
                                                Values: esriNAUFeet | esriNAUKilometers | esriNAUMeters |
                                                        esriNAUMiles | esriNAUNauticalMiles | esriNAUYards |
                                                        esriNAUUnknown
        -----------------------------------     --------------------------------------------------------------------
        directions_time_attribute_name          Optional string. The name of network attribute to use for the drive
                                                time when computing directions. The default is as defined in the network
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        output_geometry_precision               Optional float.  The precision of the output geometry after
                                                generalization. If 0, no generalization of output geometry is
                                                performed. The default is as defined in the network service
                                                configuration.
        -----------------------------------     --------------------------------------------------------------------
        output_geometry_precision_units         Optional string. The units of the output geometry precision. The
                                                default value is esriUnknownUnits.
                                                Values: esriUnknownUnits | esriCentimeters | esriDecimalDegrees |
                                                        esriDecimeters | esriFeet | esriInches | esriKilometers |
                                                        esriMeters | esriMiles | esriMillimeters |
                                                        esriNauticalMiles | esriPoints | esriYards
        -----------------------------------     --------------------------------------------------------------------
        return_z                                Optional boolean. If true, Z values will be included in the returned
                                                routes and compressed geometry if the network dataset is Z-aware.
                                                The default is false.
        ===================================     ====================================================================


        :return: dict



        """

        if not self.properties.layerType == "esriNAServerRouteLayer":
            raise ValueError("The solve operation is supported on a network "
                             "layer of Route type only")

        url = self._url + "/solve"
        params = {
                    "f" : "json",
                    "stops": stops
                 }

        if not barriers is None:
            params['barriers'] = barriers
        if not polyline_barriers is None:
            params['polylineBarriers'] = polyline_barriers
        if not polygon_barriers is None:
            params['polygonBarriers'] = polygon_barriers
        if not travel_mode is None:
            params['travelMode'] = travel_mode
        if not attribute_parameter_values is None:
            params['attributeParameterValues'] = attribute_parameter_values
        if not return_directions is None:
            params['returnDirections'] = return_directions
        if not return_routes is None:
            params['returnRoutes'] = return_routes
        if not return_stops is None:
            params['returnStops'] = return_stops
        if not return_barriers is None:
            params['returnBarriers'] = return_barriers
        if not return_polyline_barriers is None:
            params['returnPolylineBarriers'] = return_polyline_barriers
        if not return_polygon_barriers is None:
            params['returnPolygonBarriers'] = return_polygon_barriers
        if not out_sr is None:
            params['outSR'] = out_sr
        if not ignore_invalid_locations is None:
            params['ignoreInvalidLocations'] = ignore_invalid_locations
        if not output_lines is None:
            params['outputLines'] = output_lines
        if not find_best_sequence is None:
            params['findBestSequence'] = find_best_sequence
        if not preserve_first_stop is None:
            params['preserveFirstStop'] = preserve_first_stop
        if not preserve_last_stop is None:
            params['preserveLastStop'] = preserve_last_stop
        if not use_time_windows is None:
            params['useTimeWindows'] = use_time_windows
        if not start_time is None:
            params['startTime'] = start_time
        if not start_time_is_utc is None:
            params['startTimeIsUTC'] = start_time_is_utc
        if not accumulate_attribute_names is None:
            params['accumulateAttributeNames'] = accumulate_attribute_names
        if not impedance_attribute_name is None:
            params['impedanceAttributeName'] = impedance_attribute_name
        if not restriction_attribute_names is None:
            params['restrictionAttributeNames'] = restriction_attribute_names
        if not restrict_u_turns is None:
            params['restrictUTurns'] = restrict_u_turns
        if not use_hierarchy is None:
            params['useHierarchy'] = use_hierarchy
        if not directions_language is None:
            params['directionsLanguage'] = directions_language
        if not directions_output_type is None:
            params['directionsOutputType'] = directions_output_type
        if not directions_style_name is None:
            params['directionsStyleName'] = directions_style_name
        if not directions_length_units is None:
            params['directionsLengthUnits'] = directions_length_units
        if not directions_time_attribute_name is None:
            params['directionsTimeAttributeName'] = directions_time_attribute_name
        if not output_geometry_precision is None:
            params['outputGeometryPrecision'] = output_geometry_precision
        if not output_geometry_precision_units is None:
            params['outputGeometryPrecisionUnits'] = output_geometry_precision_units
        if not return_z is None:
            params['returnZ'] = return_z

        return self._con.post(path=url,
                              postdata=params, token=self._token)


class ServiceAreaLayer(NetworkLayer):
    """
    The Service Area Layer which has common properties of Network
    Layer as well as some attributes unique to Service Area Layer
    only.
    """
    def solve_service_area(self, facilities,
                           barriers=None,
                           polyline_barriers=None,
                           polygon_barriers=None,
                           travel_mode=None,
                           attribute_parameter_values=None,
                           default_breaks=None,
                           exclude_sources_from_polygons=None,
                           merge_similar_polygon_ranges=None,
                           output_lines=None,
                           output_polygons=None,
                           overlap_lines=None,
                           overlap_polygons=None,
                           split_lines_at_breaks=None,
                           split_polygons_at_breaks=None,
                           trim_outer_polygon=None,
                           trim_polygon_distance=None,
                           trim_polygon_distance_units=None,
                           return_facilities=False,
                           return_barriers=False,
                           return_polyline_barriers=False,
                           return_polygon_barriers=False,
                           out_sr=None,
                           accumulate_attribute_names=None,
                           impedance_attribute_name=None,
                           restriction_attribute_names=None,
                           restrict_u_turns=None,
                           output_geometry_precision=None,
                           output_geometry_precision_units='esriUnknownUnits',
                           use_hierarchy=None,
                           time_of_day=None,
                           time_of_day_is_utc=None,
                           travel_direction=None,
                           return_z=False):
        """ The solve service area operation is performed on a network layer
        resource of type service area (layerType is esriNAServerServiceArea).
        You can provide arguments to the solve service area operation as
        query parameters.
        Inputs:
            facilities - The set of facilities loaded as network locations
                         during analysis. Facilities can be specified using
                         a simple comma / semi-colon based syntax or as a
                         JSON structure. If facilities are not specified,
                         preloaded facilities from the map document are used
                         in the analysis. If an empty json object is passed
                         ('{}') preloaded facilities are ignored.
            barriers - The set of barriers loaded as network locations during
                       analysis. Barriers can be specified using a simple
                       comma/semicolon-based syntax or as a JSON structure.
                       If barriers are not specified, preloaded barriers from
                       the map document are used in the analysis. If an empty
                       json object is passed ('{}'), preloaded barriers are
                       ignored.
            polylineBarriers - The set of polyline barriers loaded as network
                               locations during analysis. If polyline barriers
                               are not specified, preloaded polyline barriers
                               from the map document are used in the analysis.
                               If an empty json object is passed ('{}'),
                               preloaded polyline barriers are ignored.
            polygonBarriers - The set of polygon barriers loaded as network
                              locations during analysis. If polygon barriers
                              are not specified, preloaded polygon barriers
                              from the map document are used in the analysis.
                              If an empty json object is passed ('{}'),
                              preloaded polygon barriers are ignored.
            travelMode - Travel modes provide override values that help you
                         quickly and consistently model a vehicle or mode of
                         transportation. The chosen travel mode must be
                         preconfigured on the network dataset that the
                         service area service references.
            attributeParameterValues - A set of attribute parameter values that
                                       can be parameterized to determine which
                                       network elements can be used by a vehicle.
            defaultBreaks - A comma-separated list of doubles. The default is
                            defined in the network analysis layer.
            excludeSourcesFromPolygons - A comma-separated list of string names.
                                         The default is defined in the network
                                         analysis layer.

            mergeSimilarPolygonRanges - If true, similar ranges will be merged
                                        in the result polygons. The default is
                                        defined in the network analysis layer.
            outputLines - The type of lines(s) generated. The default is as
                          defined in the network analysis layer.
            outputPolygons - The type of polygon(s) generated. The default is
                             as defined in the network analysis layer.
            overlapLines - Indicates if the lines should overlap from multiple
                           facilities. The default is defined in the network
                           analysis layer.
            overlapPolygons - Indicates if the polygons for all facilities
                              should overlap. The default is defined in the
                              network analysis layer.
            splitLinesAtBreaks - If true, lines will be split at breaks. The
                                 default is defined in the network analysis
                                 layer.
            splitPolygonsAtBreaks - If true, polygons will be split at breaks.
                                    The default is defined in the network
                                    analysis layer.
            trimOuterPolygon -  If true, the outermost polygon (at the maximum
                                break value) will be trimmed. The default is
                                defined in the network analysis layer.
            trimPolygonDistance -  If polygons are being trimmed, provides the
                                   distance to trim. The default is defined in
                                   the network analysis layer.
            trimPolygonDistanceUnits - If polygons are being trimmed, specifies
                                       the units of the trimPolygonDistance. The
                                       default is defined in the network analysis
                                       layer.
            returnFacilities - If true, facilities will be returned with the
                               analysis results. Default is false.
            returnBarriers - If true, barriers will be returned with the analysis
                             results. Default is false.
            returnPolylineBarriers - If true, polyline barriers will be returned
                                     with the analysis results. Default is false.
            returnPolygonBarriers - If true, polygon barriers will be returned
                                    with the analysis results. Default is false.
            outSR - The well-known ID of the spatial reference for the geometries
                    returned with the analysis results. If outSR is not specified,
                    the geometries are returned in the spatial reference of the map.
            accumulateAttributeNames - The list of network attribute names to be
                                       accumulated with the analysis. The default
                                       is as defined in the network analysis layer.
                                       The value should be specified as a comma
                                       separated list of attribute names. You can
                                       also specify a value of none to indicate that
                                       no network attributes should be accumulated.
            impedanceAttributeName - The network attribute name to be used as the
                                     impedance attribute in analysis. The default
                                     is as defined in the network analysis layer.
            restrictionAttributeNames - The list of network attribute names to be
                                        used as restrictions with the analysis. The
                                        default is as defined in the network analysis
                                        layer. The value should be specified as a
                                        comma separated list of attribute names.
                                        You can also specify a value of none to
                                        indicate that no network attributes should
                                        be used as restrictions.
            restrictUTurns - Specifies how U-Turns should be restricted in the
                             analysis. The default is as defined in the network
                             analysis layer. Values: esriNFSBAllowBacktrack |
                             esriNFSBAtDeadEndsOnly | esriNFSBNoBacktrack |
                             esriNFSBAtDeadEndsAndIntersections
            outputGeometryPrecision - The precision of the output geometry after
                                      generalization. If 0, no generalization of
                                      output geometry is performed. The default is
                                      as defined in the network service configuration.
            outputGeometryPrecisionUnits - The units of the output geometry precision.
                                           The default value is esriUnknownUnits.
                                           Values: esriUnknownUnits | esriCentimeters |
                                           esriDecimalDegrees | esriDecimeters |
                                           esriFeet | esriInches | esriKilometers |
                                           esriMeters | esriMiles | esriMillimeters |
                                           esriNauticalMiles | esriPoints | esriYards
            useHierarchy - If true, the hierarchy attribute for the network should be
                           used in analysis. The default is as defined in the network
                           layer. This cannot be used in conjunction with outputLines.
            timeOfDay - The date and time at the facility. If travelDirection is set
                        to esriNATravelDirectionToFacility, the timeOfDay value
                        specifies the arrival time at the facility. if travelDirection
                        is set to esriNATravelDirectionFromFacility, the timeOfDay
                        value is the departure time from the facility. The time zone
                        for timeOfDay is specified by timeOfDayIsUTC.
            timeOfDayIsUTC - The time zone or zones of the timeOfDay parameter. When
                             set to false, which is the default value, the timeOfDay
                             parameter refers to the time zone or zones in which the
                             facilities are located. Therefore, the start or end times
                             of the service areas are staggered by time zone.
            travelDirection - Options for traveling to or from the facility. The
                              default is defined in the network analysis layer.
            returnZ - If true, Z values will be included in saPolygons and saPolylines
                      geometry if the network dataset is Z-aware.
    """
        if not self.properties.layerType == "esriNAServerServiceAreaLayer":
            raise TypeError("The solveServiceArea operation is supported on a network "
                             "layer of Service Area type only")

        url = self._url + "/solveServiceArea"
        params = {
                "f" : "json",
                "facilities": facilities
                }

        if not barriers is None:
            params['barriers'] = barriers
        if not polyline_barriers is None:
            params['polylineBarriers'] = polyline_barriers
        if not polygon_barriers is None:
            params['polygonBarriers'] = polygon_barriers
        if not travel_mode is None:
            params['travelMode'] = travel_mode
        if not attribute_parameter_values is None:
            params['attributeParameterValues'] = attribute_parameter_values
        if not default_breaks is None:
            params['defaultBreaks'] = default_breaks
        if not exclude_sources_from_polygons is None:
            params['excludeSourcesFromPolygons'] = exclude_sources_from_polygons
        if not merge_similar_polygon_ranges is None:
            params['mergeSimilarPolygonRanges'] = merge_similar_polygon_ranges
        if not output_lines is None:
            params['outputLines'] = output_lines
        if not output_polygons is None:
            params['outputPolygons'] = output_polygons
        if not overlap_lines is None:
            params['overlapLines'] = overlap_lines
        if not overlap_polygons is None:
            params['overlapPolygons'] = overlap_polygons
        if not split_lines_at_breaks is None:
            params['splitLinesAtBreaks'] = split_lines_at_breaks
        if not split_polygons_at_breaks is None:
            params['splitPolygonsAtBreaks'] = split_polygons_at_breaks
        if not trim_outer_polygon is None:
            params['trimOuterPolygon'] = trim_outer_polygon
        if not trim_polygon_distance is None:
            params['trimPolygonDistance'] = trim_polygon_distance
        if not trim_polygon_distance_units is None:
            params['trimPolygonDistanceUnits'] = trim_polygon_distance_units
        if not return_facilities is None:
            params['returnFacilities'] = return_facilities
        if not return_barriers is None:
            params['returnBarriers'] = return_barriers
        if not return_polyline_barriers is None:
            params['returnPolylineBarriers'] = return_polyline_barriers
        if not return_polygon_barriers is None:
            params['returnPolygonBarriers'] = return_polygon_barriers
        if not out_sr is None:
            params['outSR'] = out_sr
        if not accumulate_attribute_names is None:
            params['accumulateAttributeNames'] = accumulate_attribute_names
        if not impedance_attribute_name is None:
            params['impedanceAttributeName'] = impedance_attribute_name
        if not restriction_attribute_names is None:
            params['restrictionAttributeNames'] = restriction_attribute_names
        if not restrict_u_turns is None:
            params['restrictUTurns'] = restrict_u_turns
        if not output_geometry_precision is None:
            params['outputGeometryPrecision'] = output_geometry_precision
        if not output_geometry_precision_units is None:
            params['outputGeometryPrecisionUnits'] = output_geometry_precision_units
        if not use_hierarchy is None:
            params['useHierarchy'] = use_hierarchy
        if not time_of_day is None:
            params['timeOfDay'] = time_of_day
        if not time_of_day_is_utc is None:
            params['timeOfDayIsUTC'] = time_of_day_is_utc
        if not travel_direction is None:
            params['travelDirection'] = travel_direction
        if not return_z is None:
            params['returnZ'] = return_z

        return self._con.post(path=url,
                              postdata=params, token=self._token)


class ClosestFacilityLayer(NetworkLayer):
    """
    The Closest Facility Network Layer which has common properties of Network
    Layer as well as some attributes unique to Closest Facility Layer
    only.
    """
    def solve_closest_facility(self, incidents, facilities,
                               barriers=None,
                               polyline_barriers=None,
                               polygon_barriers=None,
                               travel_mode=None,
                               attribute_parameter_values=None,
                               return_directions=None,
                               directions_language=None,
                               directions_style_name=None,
                               directions_length_units=None,
                               directions_time_attribute_name=None,
                               return_cf_routes=True,
                               return_facilities=False,
                               return_incidents=False,
                               return_barriers=False,
                               return_polyline_barriers=False,
                               return_polygon_barriers=False,
                               output_lines=None,
                               default_cutoff=None,
                               default_target_facility_count=None,
                               travel_direction=None,
                               out_sr=None,
                               accumulate_attribute_names=None,
                               impedance_attribute_name=None,
                               restriction_attribute_names=None,
                               restrict_u_turns=None,
                               use_hierarchy=True,
                               output_geometry_precision=None,
                               output_geometry_precision_units=None,
                               time_of_day=None,
                               time_of_day_is_utc=None,
                               time_of_day_usage=None,
                               return_z=False):
        """The solve operation is performed on a network layer resource of
        type closest facility (layerType is esriNAServerClosestFacilityLayer).
        You can provide arguments to the solve route operation as query
        parameters.
        Inputs:
            facilities  - The set of facilities loaded as network locations
                          during analysis. Facilities can be specified using
                          a simple comma / semi-colon based syntax or as a
                          JSON structure. If facilities are not specified,
                          preloaded facilities from the map document are used
                          in the analysis.
            incidents - The set of incidents loaded as network locations
                        during analysis. Incidents can be specified using
                        a simple comma / semi-colon based syntax or as a
                        JSON structure. If incidents are not specified,
                        preloaded incidents from the map document are used
                        in the analysis.
            barriers - The set of barriers loaded as network locations during
                       analysis. Barriers can be specified using a simple comma
                       / semi-colon based syntax or as a JSON structure. If
                       barriers are not specified, preloaded barriers from the
                       map document are used in the analysis. If an empty json
                       object is passed ('{}') preloaded barriers are ignored.
            polylineBarriers - The set of polyline barriers loaded as network
                               locations during analysis. If polyline barriers
                               are not specified, preloaded polyline barriers
                               from the map document are used in the analysis.
                               If an empty json object is passed ('{}')
                               preloaded polyline barriers are ignored.
            polygonBarriers - The set of polygon barriers loaded as network
                              locations during analysis. If polygon barriers
                              are not specified, preloaded polygon barriers
                              from the map document are used in the analysis.
                              If an empty json object is passed ('{}') preloaded
                              polygon barriers are ignored.
            travelMode - Travel modes provide override values that help you
                         quickly and consistently model a vehicle or mode of
                         transportation. The chosen travel mode must be
                         preconfigured on the network dataset that the routing
                         service references.
            attributeParameterValues - A set of attribute parameter values that
                                       can be parameterized to determine which
                                       network elements can be used by a vehicle.
            returnDirections - If true, directions will be generated and returned
                               with the analysis results. Default is true.
            directionsLanguage - The language to be used when computing directions.
                                 The default is as defined in the network layer. The
                                 list of supported languages can be found in REST
                                 layer description.
            directionsOutputType -  Defines content, verbosity of returned
                                    directions. The default is esriDOTStandard.
                                    Values: esriDOTComplete | esriDOTCompleteNoEvents
                                    | esriDOTInstructionsOnly | esriDOTStandard |
                                    esriDOTSummaryOnly
            directionsStyleName - The style to be used when returning the directions.
                                  The default is as defined in the network layer. The
                                  list of supported styles can be found in REST
                                  layer description.
            directionsLengthUnits - The length units to use when computing directions.
                                    The default is as defined in the network layer.
                                    Values: esriNAUFeet | esriNAUKilometers |
                                    esriNAUMeters | esriNAUMiles |
                                    esriNAUNauticalMiles | esriNAUYards |
                                    esriNAUUnknown
            directionsTimeAttributeName - The name of network attribute to use for
                                          the drive time when computing directions.
                                          The default is as defined in the network
                                          layer.
            returnCFRoutes - If true, closest facilities routes will be returned
                             with the analysis results. Default is true.
            returnFacilities -  If true, facilities  will be returned with the
                                analysis results. Default is false.
            returnIncidents - If true, incidents will be returned with the
                              analysis results. Default is false.
            returnBarriers -  If true, barriers will be returned with the analysis
                              results. Default is false.
            returnPolylineBarriers -  If true, polyline barriers will be returned
                                      with the analysis results. Default is false.
            returnPolygonBarriers - If true, polygon barriers will be returned with
                                    the analysis results. Default is false.
            outputLines - The type of output lines to be generated in the result.
                          The default is as defined in the network layer.
            defaultCutoff - The default cutoff value to stop traversing.
            defaultTargetFacilityCount - The default number of facilities to find.
            travelDirection - Options for traveling to or from the facility.
                              The default is defined in the network layer.
                              Values: esriNATravelDirectionFromFacility |
                              esriNATravelDirectionToFacility
            outSR - The spatial reference of the geometries returned with the
                    analysis results.
            accumulateAttributeNames - The list of network attribute names to be
                                       accumulated with the analysis. The default is
                                       as defined in the network layer. The value
                                       should be specified as a comma separated list
                                       of attribute names. You can also specify a
                                       value of none to indicate that no network
                                       attributes should be accumulated.
            impedanceAttributeName - The network attribute name to be used as the
                                     impedance attribute in analysis. The default is
                                     as defined in the network layer.
            restrictionAttributeNames -The list of network attribute names to be
                                       used as restrictions with the analysis. The
                                       default is as defined in the network layer.
                                       The value should be specified as a comma
                                       separated list of attribute names. You can
                                       also specify a value of none to indicate that
                                       no network attributes should be used as
                                       restrictions.
            restrictUTurns -  Specifies how U-Turns should be restricted in the
                              analysis. The default is as defined in the network
                              layer. Values: esriNFSBAllowBacktrack |
                              esriNFSBAtDeadEndsOnly | esriNFSBNoBacktrack |
                              esriNFSBAtDeadEndsAndIntersections
            useHierarchy -  If true, the hierarchy attribute for the network should
                            be used in analysis. The default is as defined in the
                            network layer.
            outputGeometryPrecision -  The precision of the output geometry after
                                       generalization. If 0, no generalization of
                                       output geometry is performed. The default is
                                       as defined in the network service
                                       configuration.
            outputGeometryPrecisionUnits - The units of the output geometry
                                           precision. The default value is
                                           esriUnknownUnits. Values: esriUnknownUnits
                                           | esriCentimeters | esriDecimalDegrees |
                                           esriDecimeters | esriFeet | esriInches |
                                           esriKilometers | esriMeters | esriMiles |
                                           esriMillimeters | esriNauticalMiles |
                                           esriPoints | esriYards
            timeOfDay - Arrival or departure date and time. Values: specified by
                        number of milliseconds since midnight Jan 1st, 1970, UTC.
            timeOfDayIsUTC - The time zone of the timeOfDay parameter. By setting
                             timeOfDayIsUTC to true, the timeOfDay parameter refers
                             to Coordinated Universal Time (UTC). Choose this option
                             if you want to find what's nearest for a specific time,
                             such as now, but aren't certain in which time zone the
                             facilities or incidents will be located.
            timeOfDayUsage - Defines the way timeOfDay value is used. The default
                             is as defined in the network layer.
                             Values: esriNATimeOfDayUseAsStartTime |
                             esriNATimeOfDayUseAsEndTime
            returnZ - If true, Z values will be included in the returned routes and
                       compressed geometry if the network dataset is Z-aware.
                       The default is false.
    """

        if not self.properties.layerType == "esriNAServerClosestFacilityLayer":
            raise TypeError("The solveClosestFacility operation is supported on a network "
                             "layer of Closest Facility type only")

        url = self._url + "/solveClosestFacility"
        params = {
                "f" : "json",
                "facilities": facilities,
                "incidents": incidents
                }

        if not barriers is None:
            params['barriers'] = barriers
        if not polyline_barriers is None:
            params['polylineBarriers'] = polyline_barriers
        if not polygon_barriers is None:
            params['polygonBarriers'] = polygon_barriers
        if not travel_mode is None:
            params['travelMode'] = travel_mode
        if not attribute_parameter_values is None:
            params['attributeParameterValues'] = attribute_parameter_values
        if not return_directions is None:
            params['returnDirections'] = return_directions
        if not directions_language is None:
            params['directionsLanguage'] = directions_language
        if not directions_style_name is None:
            params['directionsStyleName'] = directions_style_name
        if not directions_length_units is None:
            params['directionsLengthUnits'] = directions_length_units
        if not directions_time_attribute_name is None:
            params['directionsTimeAttributeName'] = directions_time_attribute_name
        if not return_cf_routes is None:
            params['returnCFRoutes'] = return_cf_routes
        if not return_facilities is None:
            params['returnFacilities'] = return_facilities
        if not return_incidents is None:
            params['returnIncidents'] = return_incidents
        if not return_barriers is None:
            params['returnBarriers'] = return_barriers
        if not return_polyline_barriers is None:
            params['returnPolylineBarriers'] = return_polyline_barriers
        if not return_polygon_barriers is None:
            params['returnPolygonBarriers'] = return_polygon_barriers
        if not output_lines is None:
            params['outputLines'] = output_lines
        if not default_cutoff is None:
            params['defaultCutoff'] = default_cutoff
        if not default_target_facility_count is None:
            params['defaultTargetFacilityCount'] = default_target_facility_count
        if not travel_direction is None:
            params['travelDirection'] = travel_direction
        if not out_sr is None:
            params['outSR'] = out_sr
        if not accumulate_attribute_names is None:
            params['accumulateAttributeNames'] = accumulate_attribute_names
        if not impedance_attribute_name is None:
            params['impedanceAttributeName'] = impedance_attribute_name
        if not restriction_attribute_names is None:
            params['restrictionAttributeNames'] = restriction_attribute_names
        if not restrict_u_turns is None:
            params['restrictUTurns'] = restrict_u_turns
        if not use_hierarchy is None:
            params['useHierarchy'] = use_hierarchy
        if not output_geometry_precision is None:
            params['outputGeometryPrecision'] = output_geometry_precision
        if not output_geometry_precision_units is None:
            params['outputGeometryPrecisionUnits'] = output_geometry_precision_units
        if not time_of_day is None:
            params['timeOfDay'] = time_of_day
        if not time_of_day_is_utc is None:
            params['timeOfDayIsUTC'] = time_of_day_is_utc
        if not time_of_day_usage is None:
            params['timeOfDayUsage'] = time_of_day_usage
        if not return_z is None:
            params['returnZ'] = return_z

        return self._con.post(path=url, postdata=params, token=self._token)


class NetworkDataset(_GISResource):
    """
    A network dataset containing a collection of network layers including route layers,
    service area layers and closest facility layers.
    """
    def __init__(self, url, gis=None):
        super(NetworkDataset, self).__init__(url, gis)
        try:
            from ..gis.server._service._adminfactory import AdminServiceGen
            self.service = AdminServiceGen(service=self, gis=gis)
        except: pass
        self._load_layers()

    @classmethod
    def fromitem(cls, item):
        """Creates a network dataset from a 'Network Analysis Service' Item in the GIS"""
        if not item.type == 'Network Analysis Service':
            raise TypeError("item must be a type of Network Analysis Service, not " + item.type)

        return cls(item.url, item._gis)

    #----------------------------------------------------------------------
    def _load_layers(self):
        """loads the various layer types"""
        self._closestFacilityLayers = []
        self._routeLayers = []
        self._serviceAreaLayers = []
        params = {
            "f" : "json",
        }
        json_dict = self._con.get(path=self._url, params=params, token=self._token)
        for k,v in json_dict.items():
            if k == "routeLayers" and json_dict[k]:
                self._routeLayers = []
                for rl in v:
                    self._routeLayers.append(
                        RouteLayer(url=self._url + "/%s" % rl,
                                   gis=self._gis))
            elif k == "serviceAreaLayers" and json_dict[k]:
                self._serviceAreaLayers = []
                for sal in v:
                    self._serviceAreaLayers.append(
                        ServiceAreaLayer(url=self._url + "/%s" % sal,
                                         gis=self._gis))
            elif k == "closestFacilityLayers" and json_dict[k]:
                self._closestFacilityLayers = []
                for cf in v:
                    self._closestFacilityLayers.append(
                        ClosestFacilityLayer(url=self._url + "/%s" % cf,
                                             gis=self._gis))
    #----------------------------------------------------------------------
    @property
    def route_layers(self):
        """List of route layers in this network dataset"""
        if self._routeLayers is None:
            self._load_layers()
        return self._routeLayers
    #----------------------------------------------------------------------
    @property
    def service_area_layers(self):
        """List of service area layers in this network dataset"""
        if self._serviceAreaLayers is None:
            self._load_layers()
        return self._serviceAreaLayers
    #----------------------------------------------------------------------
    @property
    def closest_facility_layers(self):
        """List of closest facility layers in this network dataset"""
        if self._closestFacilityLayers is None:
            self._load_layers()
        return self._closestFacilityLayers



"""
Module contains a class to manage site level functions on a local GIS
"""
from ._base import BasePortalAdmin
########################################################################
class Site(BasePortalAdmin):
    """
    Site is the root resources used after a local GIS is installed. Here
    administrators can create, export, import, and join sites.
    """
    _url = None
    _con = None
    _pa = None
    _gis = None
    _properties = None
    _json = None
    _json_dict = None
    #----------------------------------------------------------------------
    def __init__(self, url, portaladmin, **kwargs):
        """Constructor"""
        super(Site, self).__init__(url=url, gis=portaladmin._gis)
        initialize = kwargs.pop('initialize', False)
        self._url = url
        self._pa = portaladmin
        self._gis = portaladmin._gis
        self._con = portaladmin._con
        if initialize:
            self._init()
    #----------------------------------------------------------------------
    @staticmethod
    def create(con,
               url,
               username,
               password,
               full_name,
               email,
               content_store,
               description="",
               question_idx=None,
               question_ans=None):
        """
        The create site operation initializes and configures Portal for
        ArcGIS for use. It must be the first operation invoked after
        installation. Creating a new site involves:
          - Creating the initial administrator account
          - Creating a new database administrator account (which is same as
            the initial administrator account)
          - Creating token shared keys
          - Registering directories
        This operation is time consuming, as the database is initialized
        and populated with default templates and content. If the database
        directory is not empty, this operation attempts to migrate the
        database to the current version while keeping its data intact. At
        the end of this operation, the web server that hosts the API is
        restarted.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        con                             Required Connection. The connection object.
        ---------------------------     --------------------------------------------------------------------
        url                             Required string. The portal administration url
                                        Ex: https://mysite.com/<web adaptor>/portaladmin
        ---------------------------     --------------------------------------------------------------------
        username                        Required string. The initial admin account name
        ---------------------------     --------------------------------------------------------------------
        password                        Required string. The password for initial admin account
        ---------------------------     --------------------------------------------------------------------
        full_name                       Required string. The full name of the admin account
        ---------------------------     --------------------------------------------------------------------
        email                           Required string. The account email address
        ---------------------------     --------------------------------------------------------------------
        content_store                   Required string. JSON string including the path to the location of
                                        the site's content.
        ---------------------------     --------------------------------------------------------------------
        description                     Optional string. The optional description for the account
        ---------------------------     --------------------------------------------------------------------
        question_idx                    Optional integer. The index of the secret question to retrieve a
                                        forgotten password
        ---------------------------     --------------------------------------------------------------------
        question_ans                    Optional string. The answer to the secret question
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/createNewSite" % url
        params = {"f": "json",
                  "username" : username,
                  "password" : password,
                  "fullName" : full_name,
                  "email" : email,
                  "description" : description,
                  "contentStore" : content_store}
        if question_idx and question_ans:
            params['securityQuestionIdx'] = question_idx
            params['securityQuestionAns'] = question_ans
        return con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def export_site(self, location):
        """
        This operation exports the portal site configuration to a location
        you specify. The exported file includes the following information:
          Content directory - the content directory contains the data
           associated with every item in the portal
          Database dump file - a plain-text file that contains the SQL
           commands required to reconstruct the portal database
          Configuration store connection file - a JSON file that contains
           the database connection information

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        location                        Required string. The path to the folder accessible to the portal
                                        where the exported site configuration will be written.
        ===========================     ====================================================================

        :returns: dict

        .. code-block:: python

            USAGE: Export Portal Site to a location the Portal server has access to.  ** This can be a
                   lengthy operation.

            from arcgis.gis import GIS
            gis = GIS("https://yourportal.com/portal", "portaladmin", "password")
            sitemgr = gis.admin.site
            response = sitemgr.export_site(r'c:\\temp')
            print(response)

            # Output
            {'status': 'success', 'location': 'C:\\Temp\\June-9-2018-5-22-29-PM-EDT-FULL.portalsite'}

        """
        url = "%s/exportSite" % self._url
        params = {'f' : 'json',
                  'location' : location}
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def import_site(self, location):
        """
        The importSite operation lets you restore your site from a backup
        site configuration file that you created using the exportSite
        operation. It imports the site configuration file into the
        currently running portal site.
        The importSite operation will replace all site configurations with
        information included in the backup site configuration file. See the
        export_site operation documentation for details on what the backup
        file includes. The importSite operation also updates the portal
        content index.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        location                        Required string. A file path to an exported configuration.
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/importSite" % self._url
        if url.find(":7443") == -1:
            raise ValueError(
                "You must access portal not using the web adaptor (port 7443)"
            )
        params = {'f' : 'json',
                  'location' : location}
        res =  self._con.post(path=url,
                              postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return False
    #----------------------------------------------------------------------
    def join(self, admin_url, username, password):
        """
        The joinSite operation connects a portal machine to an existing
        site. You must provide an account with administrative privileges to
        the site for the operation to be successful.
        When an attempt is made to join a site, the site validates the
        administrative credentials, then returns connection information
        about its configuration store back to the portal machine. The portal
        machine then uses the connection information to work with the
        configuration store.
        If this is the first portal machine in your site, use the Create
        Site operation instead.
        The joinSite operation:
         - Registers a machine to an existing site (active machine)
         - Creates a snapshot of the database of the active machine
         - Updates the token shared key
         - Updates Web Adaptor configurations
        Sets up replication to keep the database of both machines in sync
        The operation is time-consuming as the database is configured on
        the machine and all configurations are applied from the active
        machine. After the operation is complete, the web server that hosts
        the API will be restarted.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        admin_url                       Required string. The admin URL of the existing portal site to which
                                        a machine will be joined
        ---------------------------     --------------------------------------------------------------------
        username                        Required string. The username for the initial administrator account
                                        of the existing portal site.
        ---------------------------     --------------------------------------------------------------------
        password                        Required string. The password for the initial administrator account
                                        of the existing portal site.
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/joinSite" % self._url
        params = {'f' : 'json',
                  'machineAdminUrl' : admin_url,
                  'username' : username,
                  'password' : password}
        return self._con.post(path=url,
                              postdata=params)







"""
Contains tools to manage a GIS' metadata properties
"""
from arcgis._impl.common._mixins import PropertyMap
from .. import GIS
########################################################################
class MetadataManager(object):
    """
    Provides Administrators an Easy value to enable, update and disable
    metadata settings on a Web GIS Site (Enterprise or ArcGIS Online)
    """
    _gis = None
    _portal = None
    _con = None
    #----------------------------------------------------------------------
    def __init__(self, gis):
        """Constructor"""
        self._gis = gis
        self._portal = gis._portal
        self._con = gis._portal.con
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s for %s>' % (type(self).__name__, self._gis._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._gis._url)
    #----------------------------------------------------------------------
    def enable(self, metadata_format="arcgis"):
        """
        This operation turns on metadata for items and allows the
        administrator to set the default metadata scheme.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        metadata_format                 Required string. Sets the default metadata format. The allowed
                                        values are: inspire,iso19139-3.2,fgdc,iso19139,arcgis, or iso19115
        ===========================     ====================================================================

        :returns: boolean

        """
        lookup = {
            "fgdc" : "fgdc",
            "inspire" : "inspire",
            "iso19139" : "iso19139",
            "iso19139-3.2" : "iso19139-3.2",
            "iso19115" : "iso19115",
            "arcgis" : "arcgis"
        }
        if not metadata_format.lower() in lookup.keys():
            raise ValueError("Invalid metadata_format")
        params = {
            "metadataEditable" : True,
            "metadataFormats" : lookup[metadata_format.lower()]
        }
        return self._gis.update_properties(properties_dict=params)
    #----------------------------------------------------------------------
    def disable(self):
        """
        This operation turns off metadata for items.

        :returns: boolean
        """

        params = {
            "metadataEditable" : False,
            "metadataFormats" : ""
        }
        return self._gis.update_properties(properties_dict=params)
    #----------------------------------------------------------------------
    def update(self, metadata_format="arcgis"):
        """
        This operation allows administrators to update the current metdata
        properties.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        metadata_format                 Required string. Sets the default metadata format. The allowed
                                        values are: inspire,iso19139-3.2,fgdc,iso19139,arcgis, or iso19115
        ===========================     ====================================================================

        :returns: boolean

        """
        lookup = {
            "fgdc" : "fgdc",
            "inspire" : "inspire",
            "iso19139" : "iso19139",
            "iso19139-3.2" : "iso19139-3.2",
            "iso19115" : "iso19115",
            "arcgis" : "arcgis"
        }
        if not metadata_format.lower() in lookup.keys():
            raise ValueError("Invalid metadata_format")
        params = {
            "metadataEditable" : True,
            "metadataFormats" : lookup[metadata_format.lower()]
        }
        return self._gis.update_properties(properties_dict=params)
    #----------------------------------------------------------------------
    @property
    def is_enabled(self):
        """returns boolean to show if metadata is enable on a GIS"""
        try:
            return self._gis.properties.metadataEditable
        except:
            return False



import os
import json
from ._resources import PortalResourceManager

###########################################################################
class UX(object):
    """Helper class for modifying the portal home page. This class is not created by users directly. An instance of
    the class, called 'ux', is available as a property of the GIS object. Users call methods on this 'ux' object
    to set banner, background, logo, name etc."""
    #----------------------------------------------------------------------
    def __init__(self, gis):
        """Creates helper object to manage portal home page, resources, update resources"""
        self._gis = gis
        self._portal = gis._portal
    #----------------------------------------------------------------------
    @property
    def summary(self):
        """
        Allows the get/setting of a brief summary to describe your organization on the sign in page
        associated with its custom apps and sites. This summary has a maximum of 310 characters.

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        text              Required string. The brief description of the organization.
        ================  ===============================================================

        :return: string
        """
        portal_resources = PortalResourceManager(self._gis)
        res = json.loads(open(portal_resources.get('localizedOrgProperties'), 'r').read())
        return res['default']['description']
    #----------------------------------------------------------------------

    @summary.setter
    def summary(self, text):
        """
        Allows the get/setting of a brief summary to describe your organization on the sign in page
        associated with its custom apps and sites. This summary has a maximum of 310 characters.

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        text              Required string. The brief description of the organization.
        ================  ===============================================================

        :return: string
        """
        if text == "":
            text = None
        params = {"key": "localizedOrgProperties",
                  "text": {"default":{"name": self.name,
                                      "description":text}},
                  "f": "json"
        }
        portal_resources = PortalResourceManager(self._gis)
        portal_resources.add(key='localizedOrgProperties', text=json.dumps(params['text']))
    #----------------------------------------------------------------------
    def set_banner(self, banner_file=None, is_built_in=False, custom_html = None):
        """
        Configure your home page by setting the organization's banner. You can choose one of the 5 built-in banners or
        upload your own. For best results the dimensions of the banner image should be 960 x 180 pixels. You can also
        specify a custom html for how the banner space should appear. For more information, refer to
        http://server.arcgis.com/en/portal/latest/administer/windows/configure-home.htm

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        banner_file       optional string. If uploading a custom banner, then path to the
                           banner file. If using a built-in banner, valid values are
                           banner-1, banner-2, banner-3, banner-4, banner-5. If None, existing
                           banner is remove.
        ----------------  ---------------------------------------------------------------
        is_built_in       optional bool, default=False. Specify True if using a built-in
                            banner file.
        ----------------  ---------------------------------------------------------------
        custom_html       optional string. Specify exactly how the banner should appear in
                            html. For help on this, refer to
                            http://server.arcgis.com/en/portal/latest/administer/windows/supported-html.htm
        ================  ===============================================================

        :return: True | False
        """
        #region check if banner has to be removed
        if not banner_file and not custom_html:
            #remove code
            portal_resources = PortalResourceManager(self._gis)
            #find existing banner resource file
            resource_list = portal_resources.list()
            e_banner = [banner for banner in resource_list if banner['key'].startswith('banner')]

            #loop through and remove existing banner resource file
            for banner in e_banner:
                try:
                    portal_resources.delete(banner['key'])
                except:
                    continue

            #reset the home page - recurse
            return self.set_banner('banner-2',True)
        #endregion

        #region: Set banner using banner file - built-in or new image
        if banner_file:
            rotator_panel = []
            if not is_built_in: #adding a new image file
                # find image extension
                from pathlib import Path
                fpath = Path(banner_file)
                f_splits = fpath.name.split('.')
                if len(f_splits) > 1 and f_splits[1] == 'png':
                    key_val = 'banner.png'
                elif len(f_splits) > 1 and f_splits[1] == 'jpg':
                    key_val = 'banner.jpg'
                else:
                    raise RuntimeError('Invalid image extension')

                portal_resources = PortalResourceManager(self._gis)
                add_result = portal_resources.add(key_val, banner_file)

                if add_result and custom_html:
                    rotator_panel = [{"id": "banner-custom",
                                      "innerHTML": custom_html}]

                elif add_result and not custom_html:
                    # set rotator_panel_text
                    rotator_panel = [{"id": "banner-custom",
                                      "innerHTML": "<img src='{}/portals/self/resources/{}?token=SECURITY_TOKEN' "
                                                   "style='-webkit-border-radius:0 0 10px 10px; -moz-border-radius:0 0 10px 10px;"
                                                   " -o-border-radius:0 0 10px 10px; border-radius:0 0 10px 10px; margin-top:0; "
                                                   "width:960px;'/>".format(
                                          self._portal.con.baseurl, key_val)}]
            else:  # using built-in image
                if not custom_html:  # if no custom html is specified for built-in image
                    rotator_panel = [{"id": banner_file,
                                      "innerHTML": "<img src='images/{}.jpg' "
                                                   "style='-webkit-border-radius:0 0 10px 10px; -moz-border-radius:0 0 10px 10px; "
                                                   "-o-border-radius:0 0 10px 10px; border-radius:0 0 10px 10px; margin-top:0; "
                                                   "width:960px; height:180px;'/><div style='position:absolute; bottom:80px; "
                                                   "left:80px; max-height:65px; width:660px; margin:0;'>"
                                                   "<img src='{}/portals/self/resources/thumbnail.png?token=SECURITY_TOKEN' "
                                                   "class='esriFloatLeading esriTrailingMargin025' style='margin-bottom:0; "
                                                   "max-height:100px;'/><span style='position:absolute; bottom:0; margin-bottom:0; "
                                                   "line-height:normal; font-family:HelveticaNeue,Verdana; font-weight:600; "
                                                   "font-size:32px; color:#369;'>{}</span></div>".format(banner_file,
                                                                                                         self._portal.con.baseurl,
                                                                                                         self._gis.properties.name)}]
                else:  # using custom html for built-in image
                    rotator_panel = [{"id": banner_file,
                                      "innerHTML": custom_html}]
        #endregion

        #region: Set banner just using a html text
        elif custom_html:
            rotator_panel = [{"id": "banner-html",
                              "innerHTML": custom_html}]
        #endregion

        # Update the portal self with these banner values
        update_result = self._gis.update_properties({"rotatorPanels": rotator_panel})
        return update_result
    #----------------------------------------------------------------------
    def set_logo(self, logo_file=None):
        """
        Configure your home page by setting the organization's logo image. For best results the logo file should be
        65 x 65 pixels in dimension.

        For more information, refer to http://server.arcgis.com/en/portal/latest/administer/windows/configure-general.htm

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        logo_file         optional string. Specify path to image file. If None, existing thumbnail is removed.
        ================  ===============================================================

        :return: True | False
        """

        # Add resource file
        portal_resources = PortalResourceManager(self._gis)
        key_val = ""
        # find image extension
        if logo_file is not None and \
           os.path.isfile(logo_file):
            from pathlib import Path
            fpath = Path(logo_file)
            f_splits = fpath.name.split('.')
            if len(f_splits) > 1 and f_splits[1] == 'png':
                key_val = 'thumbnail.png'
            elif len(f_splits) > 1 and f_splits[1] == 'jpg':
                key_val = 'thumbnail.jpg'
            elif len(f_splits) > 1 and f_splits[1] == 'gif':
                key_val = 'thumbnail.gif'

            add_result = portal_resources.add(key_val, logo_file)
        elif logo_file is None:
            if 'thumbnail' in dict(self._gis.properties):
                resource = self._gis.properties['thumbnail']
                if resource and len(resource) > 0:
                    portal_resources.delete(resource)
                key_val = ''
        else:
            for ext in ['.png', '.jpg', '.gif']:
                try:
                    portal_resources.delete('thumbnail' + ext)
                except:
                    continue
            key_val = None

        # Update the portal self with these banner values
        if logo_file is not None:
            update_result = self._gis.update_properties({"thumbnail": key_val})
        else:
            rp = self._gis.properties['rotatorPanels']
            for idx, r in enumerate(rp):
                if r['id'].lower() == 'banner-2':
                    r['innerHTML'] = "<img src='images/banner-2.jpg' style='-webkit-border-radius:0 0 10px 10px;" + \
                        " -moz-border-radius:0 0 10px 10px; -o-border-radius:0 0 10px 10px; border-radius:0 0 10px 10px;" +\
                        " margin-top:0; width:960px; height:180px;'/><div style='position:absolute; bottom:80px; left:80px;"+\
                        " max-height:65px; width:660px; margin:0;'><span style='position:absolute; bottom:0; " \
                        "margin-bottom:0; line-height:normal; "+\
                        "font-family:HelveticaNeue,Verdana; font-weight:600; font-size:32px; " \
                        "color:#369;'>{}</span></div>".format(self._gis.properties.name)

            update_result = self._gis.update_properties({"clearEmptyFields": True,
                                                         "thumbnail": "",
                                                         "rotatorPanels": rp})

        return update_result
    #----------------------------------------------------------------------
    def get_logo(self, download_path):
        """
        Get your organization's logo/thumbnail. You can use the `set_logo()` method to set an image as your logo.
        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        download_path     required string. Folder path to download the logo file.
        ================  ===============================================================

         :return: Path to downloaded logo file.
        """
        portal_resources = PortalResourceManager(self._gis)
        props = self._gis.properties
        if 'thumbnail' in props:
            resource = props['thumbnail']
            if resource is not None and \
               len(str(resource)) > 0:
                output = portal_resources.get(resource_name=resource,
                                              download_path=download_path)
                return output
        return None
    #----------------------------------------------------------------------
    @property
    def name(self):
        """
        Returns the site's name.  The name can get defined using the 'set_name()'.

         :return: string of the name of the site
        """
        return self._gis.properties['name']
    #----------------------------------------------------------------------
    @name.setter
    def name(self, name):
        """
        Allows for the setting of a site's name.
        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        name              required string. Name of the site.
        ================  ===============================================================


         :return: boolean
        """
        res = self._gis.update_properties({"name": name})
        if self._gis.properties.name != name:
            res = self._gis.update_properties({"name": name})
        return res
    #----------------------------------------------------------------------
    @property
    def description(self):
        """
        Returns the site's description.

         :return: dictionary
        """
        return self._gis.properties['description']
    #----------------------------------------------------------------------
    @property
    def description_visibility(self):
        """
        Returns the site's description visibility

         :return: boolean
        """
        return self._gis.properties['showHomePageDescription']
    #----------------------------------------------------------------------
    @description_visibility.setter
    def description_visibility(self, visiblity):
        """
        Allows for the setting of a site's description.
        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        visiblity         optional boolean. If True, the desciptive text will show on the
                          home page. If False, the descriptive text will not be displayed
        ================  ===============================================================

         :return: boolean
        """
        return self._gis.update_properties({'showHomePageDescription' : visiblity})
    #----------------------------------------------------------------------
    @description.setter
    def description(self, description=None):
        """
        Allows for the setting of a site's description.
        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        description       optional string. descriptive text of the site. If None, the
                          value is reset to default.
        ================  ===============================================================

         :return: boolean
        """
        if description is None:
            description = "<br/>"
        return self._gis.update_properties({'description': description})
    #----------------------------------------------------------------------
    @property
    def featured_content(self):
        """
        Returns the featured content group information.  The information
        can then be set using the 'set_featured_content()'.

          :return: dictionary

        :Usage Example:

        >>> data = ux.get_featured_content()
        >>> ux.set_featured_content(data)
        True

        """
        return {'group' : self._gis.properties['homePageFeaturedContent'],
                'count' : self._gis.properties['homePageFeaturedContentCount']}
    #----------------------------------------------------------------------
    @featured_content.setter
    def featured_content(self, content):
        """
        Sets the featured content group for the homepage.
        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        content           optional dictionary, defines the group and count of the feature
                          content area on an organizational site.  A value of None will
                          reset the value back to the install defaults.
                          Example:
                          {'group': <group id>, 'count' : 12}
        ================  ===============================================================

         :return: boolean
        """
        from .. import Group
        if content is None:
            content = {'homePageFeaturedContent': "",
                        'homePageFeaturedContentCount': 12,
                       'clearEmptyFields':True}
        elif 'group' in content and \
           isinstance(content['group'], Group):
            content['homePageFeaturedContent'] = content['group'].groupid
        elif isinstance(content, dict) and \
             'group' in content and \
             isinstance(content['group'], str):
            c = {}
            c['homePageFeaturedContent'] = content['group']
            if 'count' in content:
                c['homePageFeaturedContentCount'] = content['count']
            else:
                c['homePageFeaturedContentCount'] = 12
            content = c
        elif isinstance(content, str):
            c = {}
            c['homePageFeaturedContent'] = content
            c['homePageFeaturedContentCount'] = 12
            content = c
        self._gis.update_properties(content)
    #----------------------------------------------------------------------
    def set_background(self, background_file=None, is_built_in=True):
        """
        Configure your home page by setting the organization's background image. You can choose no image, a built-in image
        or upload your own. If you upload your own image, the image is positioned at the top and center of the page.
        The image repeats horizontally if it is smaller than the browser or device window. For best results, if you want
        a single, nonrepeating background image, the image should be 1,920 pixels wide (or smaller if your users are on
        smaller screens). The website does not resize the image. You can upload a file up to 1 MB in size.

        For more information, refer to http://server.arcgis.com/en/portal/latest/administer/windows/configure-home.htm

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        background_file   optional string. If using a custom background, specify path to image file.
                            To remove an existing background, specify None for this argument and
                            False for is_built_in argument.
        ----------------  ---------------------------------------------------------------
        is_built_in       optional bool, default=True. The built-in background is set by default.
                            If uploading a custom image, this parameter is ignored.
        ================  ===============================================================

        :return: True | False
        """

        # Add resource if using a custom background file.
        background_update_val = None
        if background_file:
            # find image extension
            from pathlib import Path
            fpath = Path(background_file)
            f_splits = fpath.name.split('.')
            if len(f_splits) > 1 and f_splits[1] == 'png':
                key_val = 'background.png'
            elif len(f_splits) > 1 and f_splits[1] == 'jpg':
                key_val = 'background.jpg'
            else:
                raise RuntimeError('Invalid image extension')

            portal_resources = PortalResourceManager(self._gis)
            add_result = portal_resources.add(key_val, background_file)
            if not add_result:
                raise RuntimeError("Error adding background image as a resource file")
            background_update_val = key_val

        elif is_built_in:  # using built-in
            background_update_val = 'images/arcgis_background.jpg'
        else:
            background_update_val = "none"

        # Update the portal self with these banner values
        update_result = self._gis.update_properties({"backgroundImage": background_update_val})
        return update_result
    #----------------------------------------------------------------------
    def get_banner(self, download_path):
        """
        Get your organization's home page banner image. You can use the `set_banner()` method to set an image or custom HTML
        code as your banner.
        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        download_path     required string. Folder path to download the banner file.
        ================  ===============================================================

         :return: Path to downloaded banner file.
        """
        #create a portal resource manager obj
        portal_resources = PortalResourceManager(self._gis)

        #find existing banner resource file
        resource_list = portal_resources.list()
        e_banner = [banner for banner in resource_list if banner['key'].startswith('banner')]

        #loop through and remove existing banner resource file
        for banner in e_banner:

            try:
                download_path = portal_resources.get(banner['key'], download_path)

            except:
                continue
        return download_path
    #----------------------------------------------------------------------
    def get_background(self, download_path):
        """
        Get your organization's home page background image. You can use the `set_background()` method to set an image
        as the home page background image.

        For more information, refer to http://server.arcgis.com/en/portal/latest/administer/windows/configure-home.htm

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        download_path     required string. Folder path to download the background file.
        ================  ===============================================================

        :return: Path to downloaded background file.
        """

        #create a portal resource manager obj
        portal_resources = PortalResourceManager(self._gis)

        #find existing banner resource file
        resource_list = portal_resources.list()
        e_background = [banner for banner in resource_list if banner['key'].startswith('background')]

        #loop through and remove existing banner resource file
        for background in e_background:

            try:
                download_path = portal_resources.get(background['key'], download_path)

            except:
                continue
        return download_path
    #----------------------------------------------------------------------
    @property
    def enable_comments(self):
        """
        Turn on item comments
        """
        return self._gis.properties['commentsEnabled']
    #----------------------------------------------------------------------
    @enable_comments.setter
    def enable_comments(self, enable=False):
        """
        Sets the comments property on the items
        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        enable            optional boolean. If True, the comments for the site are turned
                          on.  False will disable comments (default)
        ================  ===============================================================

         :return: boolean

        """
        return self._gis.update_properties({'commentsEnabled' : enable})

    #----------------------------------------------------------------------
    @property
    def default_extent(self):
        """
        returns the site's default extent

          :return: dictionary
        """
        return self._gis.properties['defaultExtent']
    #----------------------------------------------------------------------
    @default_extent.setter
    def default_extent(self, extent):
        """
        defines the site's default extent

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        extent            optional dictionary. The default extent defines where a webmap
                          will open.
                          If a value of None is given, the default extent will be provided.
                          Example Extent (default):
                          {"type":"extent","xmin":-17999999.999994524,"ymin":-11999999.999991827,
                          "xmax":17999999.999994524,"ymax":15999999.999982955,
                          "spatialReference":{"wkid":102100}}
        ================  ===============================================================

          :return: boolean
        """
        if extent is None:
            extent = {"type":"extent","xmin":-17999999.999994524,"ymin":-11999999.999991827,
                      "xmax":17999999.999994524,"ymax":15999999.999982955,
                      "spatialReference":{"wkid":102100}}
        return self._gis.update_properties({'defaultExtent' : extent})
    #----------------------------------------------------------------------
    @property
    def default_basemap(self):
        """
        returns the site's default extent

          :return: dictionary
        """
        return self._gis.properties['defaultBasemap']
    #----------------------------------------------------------------------
    @default_basemap.setter
    def default_basemap(self, basemap):
        """
        The Default Basemap opens when users click New Map. Set the group
        in the Basemap Gallery above and choose the map to open. It will
        open at the default extent you set.

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        basemap           optional string. The new default basemap to set. If None, the
                          default value will be set.
        ================  ===============================================================
        """
        if basemap is None:
            basemap = ""
        return self._gis.update_properties({'defaultBasemap' : basemap})
    #----------------------------------------------------------------------
    @property
    def vector_basemap(self):
        """
        gets/sets the default vector basemap
        """
        return self._gis.properties['defaultVectorBasemap']
    #----------------------------------------------------------------------
    @vector_basemap.setter
    def vector_basemap(self, basemap):
        """
        gets/sets the default vector basemap


        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        basemap           required dictionary. The new default vector basemap to set for
                          a given site.
        ================  ===============================================================
        """
        return self._gis.update_properties({'defaultVectorBasemap' : basemap})


"""
Updates the Federation Settings to Portal
"""
from .. import GIS
from ._base import BasePortalAdmin
########################################################################
class Federation(BasePortalAdmin):
    """
    This resource returns information about the ArcGIS Servers registered
    with Portal for ArcGIS.
    """
    _gis = None
    _url = None
    _con = None
    _portal = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        if isinstance(gis, GIS):
            url = url.replace("http://", "https://")
            self._url = url
            self._gis = gis
            self._portal = gis._portal
            self._con = gis._con
        else:
            raise ValueError("gis object must be of type GIS")
    #----------------------------------------------------------------------
    def federate(self,
                 url,
                 admin_url,
                 username,
                 password):
        """
        This operation enables ArcGIS Servers to be federated with Portal
        for ArcGIS.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        url:                            Required string. The URL of the GIS server used by external users
                                        when accessing the ArcGIS Server site. If the site includes the Web
                                        Adaptor, the URL includes the Web Adaptor address, for example,
                                        https://webadaptor.domain.com/arcgis. If you've added ArcGIS
                                        Server to your organization's reverse proxy server, the URL is
                                        the reverse proxy server address (for example,
                                        https://reverseproxy.domain.com/myorg). Note that the federation
                                        operation will perform a validation check to determine if the
                                        provided URL is accessible from the server site. If the resulting
                                        validation check fails, a warning will be generated in the Portal
                                        for ArcGIS logs. However, federation will not fail if the URL is
                                        not validated, as the URL may not be accessible from the server
                                        site, such as is the case when the server site is behind a
                                        firewall.
        ---------------------------     --------------------------------------------------------------------
        admin_url                       Required string. The URL used for accessing ArcGIS Server when
                                        performing administrative operations on the internal network, for
                                        example, https://gisserver.domain.com:6443/arcgis.
        ---------------------------     --------------------------------------------------------------------
        username                        Required string. The username of the primary site administrator account
        ---------------------------     --------------------------------------------------------------------
        password                        Required string. password of the username above.
        ===========================     ====================================================================

        :returns: dict, server response with server ID
        """
        fedurl = "%s/servers/federate" % self._url
        params = {
            "f" : "json",
            "url" : url,
            "adminUrl" : admin_url,
            "username" : username,
            "password" : password
        }
        res = self._con.post(path=fedurl,
                             postdata=params)
        if 'status' in res:
            return res['status']
        return res
    #----------------------------------------------------------------------
    @property
    def servers(self):
        """
        This resource returns detailed information about the ArcGIS Servers
        registered with Portal for ArcGIS, such as the ID of the server,
        name of the server, ArcGIS Web Adaptor URL, administration URL, and
        if the server is set as a hosting server.
        """
        url = "%s/servers" % self._url
        params = {"f" : "json"}
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    def unfederate(self, server_id):
        """
        This operation unfederates an ArcGIS Server from Portal for ArcGIS.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        server_id                       Required string. The unique ID of the server
        ===========================     ====================================================================

        :returns: boolean



        """
        url = "%s/servers/%s/unfederate" % (self._url, server_id)
        params = {"f" : "json"}
        res = self._con.post(url, params)
        if 'status' in res:
            return res['status'] == 'success'
        return False
    #----------------------------------------------------------------------
    def update(self, server_id, role, function=None):
        """
        This operation allows you to set an ArcGIS Server federated with
        Portal for ArcGIS as the hosting server or to enforce fine-grained
        access control to a federated server. You can also remove hosting
        server status from an ArcGIS Server. You can also remove hosting
        server status from an ArcGIS Server. To set a hosting server, an
        enterprise geodatabase must be registered as a managed database
        with the ArcGIS Server.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        server_id                       Required string. The unique ID of the server
        ---------------------------     --------------------------------------------------------------------
        role                            Required string. Whether the server is a hosting server for the
                                        portal, a federated server, or a server with restricted access to
                                        publishing. The allowed values are:
                                        FEDERATED_SERVER, FEDERATED_SERVER_WITH_RESTRICTED_PUBLISHING,
                                        or HOSTING_SERVER.
        ---------------------------     --------------------------------------------------------------------
        function                        Optional string. This is the purpose of the ArcGIS Server.
                                        Values are: GeoAnalytics, RasterAnalytics, ImageHosting, or None
        ===========================     ====================================================================

        :returns: boolean

        """
        role_allow = ["FEDERATED_SERVER",
                      "FEDERATED_SERVER_WITH_RESTRICTED_PUBLISHING",
                      "HOSTING_SERVER"]
        function_allow = ["GeoAnalytics",
                          "RasterAnalytics",
                          "ImageHosting"]
        if role.upper() in role_allow:
            role = role.upper()
        else:
            raise ValueError("Invalid role type")
        if function and \
           function not in function_allow:
            raise ValueError("Invalid function")
        params = {
            "f" : "json",
            "serverRole" : role,
        }
        if function:
            params["serverFunction"] = function
        url = "%s/servers/%s/update" % (self._url, server_id)
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def validate(self, server_id):
        """
        This operation provides status information about a specific ArcGIS
        Server federated with Portal for ArcGIS.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        server_id                       Required string. The unique ID of the server
        ===========================     ====================================================================

        :returns: dict

        """
        params = {"f" : "json"}
        url = "%s/servers/%s/validate" % (self._url, server_id)
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    def validate_all(self):
        """
        This operation returns information on the status of ArcGIS Servers
        registered with Portal for ArcGIS.
        """
        params = {"f" : "json"}
        url = "%s/servers/validate" % (self._url)
        return self._con.get(path=url, params=params)



import json as _json
from arcgis.raster._layer import ImageryLayer as _ImageryLayer
import arcgis as _arcgis
import string as _string
import random as _random


def _set_context(params, function_context = None):
    out_sr = _arcgis.env.out_spatial_reference
    process_sr = _arcgis.env.process_spatial_reference
    out_extent = _arcgis.env.analysis_extent
    mask = _arcgis.env.mask
    snap_raster = _arcgis.env.snap_raster
    cell_size = _arcgis.env.cell_size
    parallel_processing_factor = _arcgis.env.parallel_processing_factor

    context = {}

    if out_sr is not None:
        context['outSR'] = {'wkid': int(out_sr)}

    if out_extent is not None:
        context['extent'] = out_extent

    if process_sr is not None:
        context['processSR'] = {'wkid': int(process_sr)}


    if mask is not None:
        if isinstance(mask, _ImageryLayer):
            context['mask'] = {"url":mask._url}
        elif isinstance(mask,str):
            context['mask'] = {"url":mask}
    
    if cell_size is not None:
        if isinstance(cell_size, _ImageryLayer):
            context['cellSize'] = {"url":cell_size._url}
        elif isinstance(cell_size,str):
            if 'http:' in cell_size or 'https:' in cell_size:
                context['cellSize'] = {"url":cell_size}
            else:
                context['cellSize'] = cell_size
        else:
            context['cellSize'] = cell_size

    if snap_raster is not None:
        if isinstance(snap_raster, _ImageryLayer):
            context['snapRaster'] = {"url":snap_raster._url}
        elif isinstance(mask,str):
            context['snapRaster'] = {"url":snap_raster}


    if parallel_processing_factor is not None:
        context['parallelProcessingFactor'] = parallel_processing_factor


    if function_context is not None:
        if context is not None:
            context.update({k: function_context[k] for k in function_context.keys()})

        else:
            context = function_context

    if context:
        params["context"] = _json.dumps(context)

def _id_generator(size=6, chars=_string.ascii_uppercase + _string.digits):
    return ''.join(_random.choice(chars) for _ in range(size))



"""
arcgis2geojson is a derivative work of ESRI's arcgis-to-geojson-utils:
https://github.com/Esri/arcgis-to-geojson-utils/
Original code is Copyright 2015 by Esri and was licensed under
the Apache License, Version 2.0: http://www.apache.org/licenses/LICENSE-2.0
Ported to Python in 2016 by Chris Shaw.

arcgis2geojson is made available under the MIT License.
"""

import numbers


def pointsEqual(a, b):
    """
    checks if 2 [x, y] points are equal
    """
    for i in range(0, len(a)):
        if a[i] != b[i]:
            return False
    return True


def closeRing(coordinates):
    """
    checks if the first and last points of a ring are equal and closes the ring
    """
    if not pointsEqual(coordinates[0], coordinates[len(coordinates) - 1]):
        coordinates.append(coordinates[0])
    return coordinates


def ringIsClockwise(ringToTest):
    """
    determine if polygon ring coordinates are clockwise. clockwise signifies
    outer ring, counter-clockwise an inner ring or hole.
    """

    total = 0
    i = 0
    rLength = len(ringToTest)
    pt1 = ringToTest[i]
    pt2 = None
    for i in range(0, rLength - 1):
        pt2 = ringToTest[i + 1]
        total += (pt2[0] - pt1[0]) * (pt2[1] + pt1[1])
        pt1 = pt2

    return (total >= 0)


def vertexIntersectsVertex(a1, a2, b1, b2):
    uaT = (b2[0] - b1[0]) * (a1[1] - b1[1]) - (b2[1] - b1[1]) * (a1[0] - b1[0])
    ubT = (a2[0] - a1[0]) * (a1[1] - b1[1]) - (a2[1] - a1[1]) * (a1[0] - b1[0])
    uB = (b2[1] - b1[1]) * (a2[0] - a1[0]) - (b2[0] - b1[0]) * (a2[1] - a1[1])

    if uB != 0:
        ua = uaT / uB
        ub = ubT / uB

        if ua >= 0 and ua <= 1 and ub >= 0 and ub <= 1:
            return True

    return False


def arrayIntersectsArray(a, b):
    for i in range(0, len(a)-1):
        for j in range(0, len(b)-1):
            if vertexIntersectsVertex(a[i], a[i + 1], b[j], b[j + 1]):
                return True

    return False


def coordinatesContainPoint(coordinates, point):

    contains = False
    l = len(coordinates)
    i = -1
    j = l - 1
    while ((i + 1) < l):
        i = i + 1
        ci = coordinates[i]
        cj = coordinates[j]
        if ((ci[1] <= point[1] and point[1] < cj[1]) or (cj[1] <= point[1] and point[1] < ci[1])) and\
           (point[0] < (cj[0] - ci[0]) * (point[1] - ci[1]) / (cj[1] - ci[1]) + ci[0]):
            contains = not contains
        j = i
    return contains


def coordinatesContainCoordinates(outer, inner):
    intersects = arrayIntersectsArray(outer, inner)
    contains = coordinatesContainPoint(outer, inner[0])
    if not intersects and contains:
        return True
    return False


def convertRingsToGeoJSON(rings):
    """
    do any polygons in this array contain any other polygons in this array?
    used for checking for holes in arcgis rings
    """

    outerRings = []
    holes = []
    x = None  # iterator
    outerRing = None  # current outer ring being evaluated
    hole = None  # current hole being evaluated

    # for each ring
    for r in range(0, len(rings)):
        ring = closeRing(rings[r])
        if len(ring) < 4:
            continue

        # is this ring an outer ring? is it clockwise?
        if ringIsClockwise(ring):
            polygon = [ring]
            outerRings.append(polygon)  # push to outer rings
        else:
            holes.append(ring)  # counterclockwise push to holes

    uncontainedHoles = []

    # while there are holes left...
    while len(holes):
        # pop a hole off out stack
        hole = holes.pop()

        # loop over all outer rings and see if they contain our hole.
        contained = False
        x = len(outerRings) - 1
        while (x >= 0):
            outerRing = outerRings[x][0]
            if coordinatesContainCoordinates(outerRing, hole):
                # the hole is contained push it into our polygon
                outerRings[x].append(hole)
                contained = True
                break
            x = x-1

        # ring is not contained in any outer ring
        # sometimes this happens https://github.com/Esri/esri-leaflet/issues/320
        if not contained:
            uncontainedHoles.append(hole)

    # if we couldn't match any holes using contains we can try intersects...
    while len(uncontainedHoles):
        # pop a hole off out stack
        hole = uncontainedHoles.pop()

        # loop over all outer rings and see if any intersect our hole.
        intersects = False
        x = len(outerRings) - 1
        while (x >= 0):
            outerRing = outerRings[x][0]
            if arrayIntersectsArray(outerRing, hole):
                # the hole is contained push it into our polygon
                outerRings[x].append(hole)
                intersects = True
                break
            x = x-1

        if not intersects:
            outerRings.append([hole[::-1]])

    if len(outerRings) == 1:
        return {
            'type': 'Polygon',
            'coordinates': outerRings[0]
        }
    else:
        return {
            'type': 'MultiPolygon',
            'coordinates': outerRings
        }


def arcgis2geojson(arcgis, idAttribute=None):
    """
    Convert an ArcGIS JSON object to a GeoJSON object
    """

    geojson = {}

    if 'x' in arcgis and isinstance(arcgis['x'], numbers.Number) and\
        'y' in arcgis and isinstance(arcgis['y'], numbers.Number):
        geojson['type'] = 'Point'
        geojson['coordinates'] = [arcgis['x'], arcgis['y']]

    if 'points' in arcgis:
        geojson['type'] = 'MultiPoint'
        geojson['coordinates'] = arcgis['points']

    if 'paths' in arcgis:
        if len(arcgis['paths']) == 1:
            geojson['type'] = 'LineString'
            geojson['coordinates'] = arcgis['paths'][0]
        else:
            geojson['type'] = 'MultiLineString'
            geojson['coordinates'] = arcgis['paths']

    if 'rings' in arcgis:
        geojson = convertRingsToGeoJSON(arcgis['rings'])

    if 'geometry' in arcgis or 'attributes' in arcgis:
        geojson['type'] = 'Feature'
        if 'geometry' in arcgis:
            geojson['geometry'] = arcgis2geojson(arcgis['geometry'])
        else:
            geojson['geometry'] = None

        if 'attributes' in arcgis:
            geojson['properties'] = arcgis['attributes']
            if idAttribute in arcgis['attributes']:
                geojson['id'] = arcgis['attributes'][idAttribute]
            elif 'OBJECTID' in arcgis['attributes']:
                geojson['id'] = arcgis['attributes']['OBJECTID']
            elif 'FID' in arcgis['attributes']:
                geojson['id'] = arcgis['attributes']['FID']
        else:
            geojson['properties'] = None

    if 'geometry' in geojson and not(geojson['geometry']):
        geojson['geometry'] = None

    return geojson



"""
The arcgis.widgets module provides components for visualizing GIS data and analysis.
This module includes the MapView Jupyter notebook widget for visualizing maps and layers
"""
import json
import random
import string
import time
import logging
import base64
import urllib.request
from uuid import uuid4
from collections import OrderedDict

from ipywidgets import widgets
from ipywidgets.embed import embed_minimal_html
from traitlets import Unicode, Int, List, Bool, Dict, Tuple, Float, observe
from IPython.display import display, HTML

from arcgis.widgets._webscene_utils import DEFAULT_WEBSCENE_TEXT_PROPERTY
from arcgis.widgets._loading_icon_str import _loading_icon_str
from arcgis import __version__ as py_api_version
import arcgis.mapping
import arcgis

log = logging.getLogger(__name__)

DEFAULT_ELEMENT_HEIGHT = "400px"

_DEFAULT_JS_CDN = "https://js.arcgis.com/4.8/"
_js_cdn_override_global = ""

def _is_iterable(obj):
    """Returns true for all iterables but strings"""
    try:
        _ = iter(obj)
        return not isinstance(obj, str)
    except TypeError:
        return False

def _make_jsonable_dict(obj):
    """Attempts to make any object a jsonable dict. Will exclude any
    value that is non-serializable"""
    flag = "K9SS8NRqLyiXF6pXDELETEME"
    def default_func(x):
        """If you come across anything non-jsonable, return a flag that
        will later be used to delete all values with this value"""
        return flag
    dict_ = json.loads(json.dumps(obj, default=default_func))
    return { k:v for k, v in dict_.items() if v != flag }

@widgets.register
class MapView(widgets.DOMWidget):
    """Mapping widget for Jupyter Notebook and JupyterLab.

    ==================     ====================================================================
    **Argument**              **Description**
    ------------------     --------------------------------------------------------------------
    gis                    The active GIS() instance you want this map widget to use.
    ------------------     --------------------------------------------------------------------
    item                   A `WebMap` or `WebScene` item instance that you want to visualize
    ------------------     --------------------------------------------------------------------
    mode                   Whether to construct a '2D' map or '3D' map. See the `mode` property
                           for more information.
    ==================     ====================================================================

    """

    # region Class, instance and interop variables
    _view_name = Unicode('ArcGISMapIPyWidgetView').tag(sync=True)
    _model_name = Unicode('ArcGISMapIPyWidgetModel').tag(sync=True)
    _view_module = Unicode('arcgis-map-ipywidget').tag(sync=True)
    _model_module = Unicode('arcgis-map-ipywidget').tag(sync=True)
    _view_module_version = Unicode(str(py_api_version)).tag(sync=True)
    _model_module_version = Unicode(str(py_api_version)).tag(sync=True)

    # Start model specific state
    # Start map specific drawing state
    zoom = Float(2).tag(sync=True)
    """What level of zoom you want to apply: the higher the number, the more
    zoomed in you are.
    """
    rotation = Float(0).tag(sync=True)
    """For 2D mode, the clockwise rotation of due north in relation to the top
    of the view in degrees. Note that you can NOT set rotation in 3D mode.
    3D mode uses the ‘heading’ property.
    """
    heading = Float(0).tag(sync=True)
    """For 3D mode, the compass heading of the camera in degrees. Heading is
    zero when north is the top of the screen. It increases as the view rotates
    clockwise. The angles are always normalized between 0 and 360 degrees.
    Note that you can NOT set heading in 2D mode. 2D mode uses the ‘rotation’
    property.
    """
    tilt = Float(0).tag(sync=True)
    """For 3D mode, the tilt of the camera in degrees with respect to the
    surface as projected down from the camera position. Tilt is zero when
    looking straight down at the surface and 90 degrees when the camera is
    looking parallel to the surface. Note that you can NOT set tilt in
    2D mode.
    """

    @property
    def basemap(self):
        """What basemap you would like to apply to the widget (‘topo’,
        ‘national-geographic’, etc.). See `basemaps` for a full list
        """
        return self._basemap

    @basemap.setter
    def basemap(self, value):
        if value in self.basemaps:
            self._basemap = value
        elif value in self.gallery_basemaps:
            self._basemap = value
        else:
            raise RuntimeError("Basemap '{}' isn't valid".format(value))

    _basemap = Unicode('topo').tag(sync=True)
    """What basemap you would like to apply to the widget (‘topo’,
    ‘national-geographic’, etc.). See `basemaps` for a full list
    """
    mode = Unicode('2D').tag(sync=True)
    """The string that specifies whether the map displays in '2D' mode
    (MapView) or '3D' mode (SceneView). Possible values: '2D', '3D'.

    .. raw:: html

        <p>Note that you can also toggle between '2D' and
        '3D' mode by pressing the
        <img src=" data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAA
        Bzenr0AAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4
        gcFFxIGmlhcGAAAAslJREFUWMPt10+IV1UUB/DPNMMYCTqzqV4GLkpCkIwDLgYKbGG4iLIa
        KNNdRoGFUIFFUpYtLNBVhFIQJJQKEaK5CIpgWgxmZxVRWJsIHiSNM4WG5mCbG/x4zPzmzc8
        ZKvDs7r3nvvO958/3nMc1+Zelr3MRERdxZo47K3AOF7AqM5dcDYCBxvpMZq6ZTTkinsAWPJ
        SZUxHx7dV6YKCtYkSsxsPYkJnTCxWC6+ahey8+WkjjrT0QEbdiO+5c6CQcaGG8HxtxfLbXR
        8ThGbZHMnNlTwAiog+bcQrf4BeMzPaRzHysJajWHliBvfi6ABnHXxGxEY9gJyYXMwQbcCNO
        Z+bJ8qKdeAA/4DsMYvmiAMjM9yNi+T/nJSSPZ+baRph+bLj9JXyVmWMde0PYk5nPzrcMa2y
        JiEG8jM8bIK/gz8advVgXEfcU48PYjV298MAkJnAZv2F1KUddkvFKZu7H2lKyu/BKZk617Q
        XzptbZqDsiHsWJzDz//+mGXchoKe7PzCNz6A3iSWzFUAnfBI7gvcy81LzT38L4MryGbVVV9
        dV1faoLXR/D97ipcMWlkoT3YbSqqtN1Xf/RGkApxdexB+twvKqq0aqqxuu67tRbgpPYgZ8R
        uB4/dZT629hXVdWJuq6n23bDN7A7M8+VhBsr7PhiQ+8pTJdXj+KtzHwTR0sXHcrMs/gE21o
        3o5nIo4AYa2xvLS881ii5o/gS5wsnfIrP8E4v80A3uaGEaFlHWDbhjrK8BYexCit7mojmkH
        68Whjz+TI9HcJtmflrB6i7MdwrgJGZWmxpxb8XL/QXGr4ZZ9EkobtKI5s/gDmGiw/xXEnOf
        XgG6/FFRExgc2ZOloQ8tBghOFBmhhcwXhrV9uL2D3A5ItaU+L/b61DazTsX8WAZVtZHxO0d
        x8PlbD82NdmwbyF5vRDS0+XfYQhT5QfmYxwsQK/Jf0v+BvBD+VardrUKAAAAAElFTkSuQmC
        C"> icon in the widget UI.</p>

    """

    _readonly_extent = Dict({}).tag(sync=True)
    _extent = Dict({}).tag(sync=True)

    @property
    def extent(self):
        """A property representing the map widget's extent.


        ==================     ====================================================================
        getter                 A dict that represents the JSON of the map widget's extent.
        ------------------     --------------------------------------------------------------------
        setter                 A [[xmin, ymin], [xmax, ymax]] list, Spatially Enabled Data Frame `full_extent`,
                               or a dict that represents the JSON of the map widget's extent.

                               Examples for each:
                               web_map.extent = [[-124.35, 32.54], [-114.31, 41.95]]
                               web_map.extent = data_frame.spatial.full_extent
                               web_map.extent = {
                                    "xmin": -124.35,
                                    "ymin": 32.54,
                                    "xmax": -114.31,
                                    "ymax": 41.95
                                }
        ==================     ====================================================================

        """
        if self._readonly_extent:
            return self._readonly_extent
        else:
            return self._extent
    @extent.setter
    def extent(self, value):
        try:
            if isinstance(value, dict):
                self._extent = value
            elif _is_iterable(value) and isinstance(value, tuple):
                self._extent = {
                    "xmin": value[0],
                    "ymin": value[1],
                    "xmax": value[2],
                    "ymax": value[3]
                }
            elif _is_iterable(value):
                self._extent = {
                    "xmin": value[0][0],
                    "ymin": value[0][1],
                    "xmax": value[1][0],
                    "ymax": value[1][1]}
            else:
                raise Exception
        except Exception:
            log.warn("extent must be set to either a 2d list, spatially " \
                     "enabled data frame full_extent, or dict. Values specified " \
                "must include xmin, ymin, xmax, ymax. Please see the API doc for " \
                "more information")

    _readonly_center = Dict({}).tag(sync=True)
    _center = Dict({}).tag(sync=True)
    _center_long_lat = List([]).tag(sync=True)

    @property
    def center(self):
        """
        A property that represents the map widget's center.

        ==================     ====================================================================
        getter                 A dict that represents the JSON of the map widget's center.
        ------------------     --------------------------------------------------------------------
        setter                 A [lat, long] list, or a dict that represents the JSON of the map
                               widget's center.
        ==================     ====================================================================

        """
        if self._readonly_center:
            return self._readonly_center
        else:
            return self._center
    @center.setter
    def center(self, value):
        if isinstance(value, dict):
            self._center = value
        elif len(value) != 2:
            log.warn("If setting center to a list/tuple, the len() "\
                     "must be exactly 2 entries long")
        else:
            self._center_long_lat = [value[1], value[0]]

    # Start layer specific state
    # There is no backbone event type for a foolist.append(), but writing a new
    # tuple to self.portal_items will trigger a model change
    _add_this_notype_layer = Dict({}).tag(sync=True)
    _draw_these_notype_layers_on_widget_load = Tuple(tuple()).tag(sync=True)

    _layers_to_remove = Tuple(()).tag(sync=True)
    _add_this_graphic = Dict({}).tag(sync=True)
    _draw_these_graphics_on_widget_load = Tuple(tuple()).tag(sync=True)

    # End layer specific state
    # Start webmap/webscene state
    _webmap = Dict({}).tag(sync=True)
    _webscene = Dict({}).tag(sync=True)
    _trigger_webscene_save_to_this_portal_id = Unicode('').tag(sync=True)
    _readonly_webmap_from_js = Dict({}).tag(sync=True)
    # end webmap/webscene state
    # start miscellanous model state
    _custom_msg = Unicode('').tag(sync=True)
    _portal_token = Unicode('').tag(sync=True)
    _auth_mode = Unicode('').tag(sync=True)
    _portal_url = Unicode('').tag(sync=True)
    _portal_sharing_rest_url = Unicode('').tag(sync=True)
    _username = Unicode('').tag(sync=True)
    _trigger_interactive_draw_mode_for = Unicode('').tag(sync=True)
    _trigger_new_jlab_window_with_args = Dict({}).tag(sync=True)
    hide_mode_switch = Bool(False).tag(sync=True)
    """When set to 'True' will hide the 2D/3D switch button from the widget.
    Note that once the button is hidden, it cannot be made visible again:
    you have to create a new MapView instance.
    """
    jupyter_target = Unicode('').tag(sync=True)
    """A readonly string that is either 'lab' or 'notebook': represents if
    this widget is drawn in a Jupyter Notebook environment, or a JupyterLab
    environment
    """
    tab_mode = Unicode('auto').tag(sync=True)
    """
    .. raw:: html

        <p>This string property specifies the 'default' behavior of toggling a
        new window in a JupyterLab environment, whether that is called by
        pressing the <img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAC
        AAAAAgCAYAAABzenr0AAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAAALEwEAmpwYA
        AAAB3RJTUUH4gcUAAIIxu7mQQAAAB1pVFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoI
        EdJTVBkLmUHAAABI0lEQVRYw+1Wy3GDMBTch3JLBanA+TQR0gDz4JB0Y1xISkBAA4Eq
        HOeYU6qAzYVkPAyWYSLii/Ymzb63q89qBAQEXBgyNamquYhs5zSw1v72SNOUp3gkd2V
        Z5uP5K1dzkrslqxnzRWQD4MVV4zQw5XguP8uyDckGwBeAm1M10Rrnqqq3fd+3JEVEYh
        c3WkMcQDMM46IoPv7NQJIkdyLS/oiXZXk4VxP5FDfGNEvEz17Cuciy7J7kGwB0XRfXd
        X0Yp+NoZ/waGMQbABzE35ekKfIkDhF5mhJf7Q6o6sOReFwUxd7BzdM0ffR5BALglaQA
        iK21eydZZEsSAFpfBkjy2Rhz7Vr5qimoqurTR4JWeYqDgWDAWwpUNb+ogbn/woCAgL/
        gG9knfWwKBGcvAAAAAElFTkSuQmCC"> icon in the widget UI, or by calling
        <i>toggle_window_view()</i> function without arguments.</p>

    Note that after a widget is 'seperated' from the notebook, you can drag
    it, split it, put it in a new tab, etc. See the JupyterLab guide pages
    for more information.

    ==================     ====================================================================
    **Value**              **Description**
    ------------------     --------------------------------------------------------------------
    'auto'                 The default tab_mode: will move the widget to a `split-right` view
                           if there are no other open windows besides the current notebook,
                           otherwise will place the widget in a new tab.
    ------------------     --------------------------------------------------------------------
    'split-top'            Will move the widget above the current notebook instance.
    ------------------     --------------------------------------------------------------------
    'split-bottom'         Will move the widget below the current notebook instance.
    ------------------     --------------------------------------------------------------------
    'split-left'           Will move the widget to the left of the current notebook instance.
    ------------------     --------------------------------------------------------------------
    'split-right'          Will move the widget to the right of the current notebook instance.
    ------------------     --------------------------------------------------------------------
    'tab-before'           Will move the widget to a new tab 'before' the tab that represents
                           the current notebook instance.
    ------------------     --------------------------------------------------------------------
    'tab-after'            Will move the widget to a new tab 'after' the tab that represents
                           the current notebook instance.
    ==================     ====================================================================
    """

    ready = Bool(False).tag(sync=True)
    """A readonly bool that represents if the map widget has been drawn
    in the notebook
    """
    _js_cdn_override = Unicode().tag(sync=True)

    legend = Bool(False).tag(sync=True)
    """If set to `True`, will display a legend in the widget that will
    describe all layers added to the map. If set to `False`, will hide the
    legend. Default: `False`.
    """

    _gallery_basemaps = Dict({}).tag(sync=True)

    @property
    def gallery_basemaps(self):
        """
        View your portal's custom basemap group
        """
        if len(self._gallery_basemaps) <= 1:
            # If the only loaded gallery_basemaps is 'default', load the rest
            bmquery = self.gis.properties['basemapGalleryGroupQuery']
            basemapsgrp = self.gis.groups.search(bmquery, outside_org=True)
            if len(basemapsgrp) == 1:
                for bm in basemapsgrp[0].content():
                    if bm.type.lower() == 'web map': #  Only use WebMaps
                        item_data = bm.get_data()
                        bm_title = bm.title.lower().replace(" ", "_")
                        self._gallery_basemaps[bm_title] = item_data['baseMap']
                # Appending to dict doesn't cause change in model:
                # Must overwrite with a blank dict, then put the new dict
                copy_gallery = dict(self._gallery_basemaps)
                self._gallery_basemaps = {}
                self._gallery_basemaps = copy_gallery
                return list(self._gallery_basemaps.keys())
            else:
                return list(self._gallery_basemaps.keys())
        else:
            return list(self._gallery_basemaps.keys())

    _uuid = Unicode("").tag(sync=True)

    # end miscellanous model state
    # End model specific state

    # Start Other properties that don't interact with the model
    _hashed_layers = OrderedDict() # how we store layers
    _default_webscene_text_property = DEFAULT_WEBSCENE_TEXT_PROPERTY

    @property
    def layers(self):
        """A list of the JSON representation of layers added to the map widget
        using the add_layers() method
        """
        return [self._hashed_layers[key] for key in self._hashed_layers]
    @layers.setter
    def layers(self, value):
        raise Exception("Can't set layers directly: use add_layer()")
    # end how we store layers

    basemaps = ['dark-gray',
                'dark-gray-vector',
                'gray',
                'gray-vector',
                'hybrid',
                'national-geographic',
                'oceans',
                'osm',
                'satellite',
                'streets',
                'streets-navigation-vector',
                'streets-night-vector',
                'streets-relief-vector',
                'streets-vector',
                'terrain',
                'topo',
                'topo-vector']
    """A list of possible basemaps to set `.basemap` with
    """
    # End other properties that don't interact with the model

    def __init__(self, gis=None, item=None, mode="2D",**kwargs):
        """Constructor of Map widget.
        Accepts the following keyword arguments:
        gis     The gis instance with which the map widget works, used for authentication, and adding secure layers and
                private items from that GIS
        item    web map item from portal with which to initialize the map widget
        """
        super(MapView, self).__init__(**kwargs)
        self._uuid = str(uuid4())

        # Set up the visual display of the layout
        self.layout.height = DEFAULT_ELEMENT_HEIGHT
        self.layout.width = "100%"

        # Set up gis object
        gis = arcgis.env.active_gis if gis is None else gis
        if gis:
            self._setup_gis_properties(gis)
        else:
            from arcgis.gis import GIS
            gis = GIS(set_active=False)
            self._setup_gis_properties(gis)
            self._auth_mode = "anonymous"

        # Set up miscellanous properties needed on startup
        self.mode = mode
        self._hashed_layers = OrderedDict()
        self._setup_js_cdn()
        self._setup_default_basemap()

        # Handle webmaps and webscenes
        self.webmap_item = None
        self.webscene_item = None
        if item != None:
            self._check_if_webmap(item)
            self._check_if_webscene(item)
        else:
            from arcgis.mapping import WebMap
            self.webmap = WebMap()

        # Handle callbacks and such
        self.on_msg(self._handle_map_msg)
        self._draw_end_handlers = widgets.CallbackDispatcher()
        self._click_handlers = widgets.CallbackDispatcher()

    # Start screenshot specific section

    def _ipython_display_(self):
        """Override the parent ipython display function that is called
        whenever is displayed in the notebook. Display a blank area
        below the map widget that can be controlled via a display handler
        set to self._preview_image_display_handler.
        """
        super(MapView, self)._ipython_display_()
        self._preview_image_display_handler = display(
            HTML(self._assemble_img_preview_html_str("")),
            display_id = "preview-" + str(self._uuid))

    def _assemble_img_preview_html_str(self, img_src):
        """Helper function that creates an HTML string of the <img> tag
        to add to the notebook with the correct <div> class to be hidden
        """
        img_html = '<img src="' + img_src + '"></img>'
        class_id = 'map-static-img-preview-' + self._uuid
        return '<div class="' + class_id + '">' + img_html + '</div>'

    _preview_screenshot_callback_resp = Unicode("").tag(sync=True)

    @observe('_preview_screenshot_callback_resp')
    def _preview_image_update_callback(self, change):
        """Called every time the front end takes a screenshot for a
        map preview
        """
        img_data_uri_str = self._parse_js_resp(change['new'])
        self._preview_image_display_handler.update(
           HTML(self._assemble_img_preview_html_str(img_data_uri_str)))

    _cell_output_screenshot_callback_resp = Unicode("").tag(sync=True)

    @observe('_cell_output_screenshot_callback_resp')
    def _cell_output_screenshot_update_callback(self, change):
        """Called every time the front end takes a screenshot for a 
        cell output"""
        if self._cell_output_display_handler:
            img_data_uri_str = self._parse_js_resp(change['new'])
            self._cell_output_display_handler.update(
                HTML("<img src=" + img_data_uri_str + "></img>"))

    _file_output_screenshot_callback_resp = Unicode("").tag(sync=True)

    @observe('_file_output_screenshot_callback_resp')
    def _file_output_screenshot_update_callback(self, change):
        """Called every time the front end takes a screenshot for a file
        write"""
        if self._screenshot_file_output_path:
            img_data_uri_str = self._parse_js_resp(change['new'])
            img_data_raw_str = img_data_uri_str.split('base64,')[-1]
            with open(self._screenshot_file_output_path, "wb") as f:
                img_data_raw_bytes = base64.b64decode(img_data_raw_str)
                f.write(img_data_raw_bytes)


    def _parse_js_resp(self, change_new):
        """In 3D mode, the Data URI is returned from JS. In 2D mode, a URL
        to a png is returned. This function will always return a base64
        data URI of the image represented
        """
        if not change_new.startswith("http"):
            # Already is a data URI
            return change_new
        else:
            # It's a URL: download and parse to base64 data URI, return it
            with urllib.request.urlopen(change_new) as resp:
                encoded_body = base64.b64encode(resp.read())
                return 'data:image/png;base64,{}'.format(encoded_body.decode())

    print_service_url = Unicode("").tag(sync=True)
    """The print service URL used when taking 2D screenshots. This member will
    be auto populated with the value from
    `gis.properties.helperServices.printTask['url']`. If it can't find that 
    value, it will use the default value of
    `"https://utility.arcgisonline.com/arcgis/rest/services/Utilities/PrintingTools/GPServer/Export%20Web%20Map%20Task"`.

    You can set this value to a valid Export Web Map Task URL, which will
    be used as the override for all future `take_screenshot()` calls in 2D
    mode.
    """

    def _setup_screenshot_print_service_url(self):
        """If using anon GIS, use the public Export Web Map Task. If portal,
        try to use the portal's Export Web Map Task. If you can't find the
        portal's Export Web Map task, fallback to AGOL's service
        """
        default_print_service_url = 'https://utility.arcgisonline.com/arcgis'\
            '/rest/services/Utilities/PrintingTools/GPServer'\
            '/Export%20Web%20Map%20Task'
        try:
            self.print_service_url = \
                self.gis.properties.helperServices.printTask['url']
        except Exception:
            self.print_service_url = default_print_service_url

    _trigger_screenshot_with_args = Dict({}).tag(sync=True)

    def take_screenshot(self, output_in_cell=True, set_as_preview=True,
                        file_path = ""):
        """Takes a screenshot of the current widget view. Only works in a 
        Jupyter Notebook environment.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        output_in_cell         Optional bool, default `True`. Will display the screenshot in the
                               output area of the cell where this function is called.
        ------------------     --------------------------------------------------------------------
        set_as_preview         Optional bool, default `True`. Will set the screenshot as the static
                               image preview in the cell where the map widget is being displayed.
                               Use this flag if you want the generated HTML previews of your 
                               notebook to have a map image visible.
        ------------------     --------------------------------------------------------------------
        file_path              Optional String, default `""`. To output the screenshot to a `.png`
                               file on disk, set this String to a writeable location file path
                               (Ex. `file_path = "/path/to/documents/my_screenshot.png"`).
        ==================     ====================================================================

        In all notebook outputs, each image will be encoded to a base64
        data URI and wrapped in an HTML <img> tag, like
        `<img src="base64Str">`. This means that the data for the image lives
        inside the notebook file itself, allowing for easy sharing of 
        notebooks and generated HTML previews of notebooks.

        .. note::
            This function acts asyncronously, meaning that the Python function
            will return right away, with the notebook outputs/files being 
            written after an indeterminate amount of time. Avoid calling this 
            function  multiple times in a row if the asyncronous portion of 
            the function hasn't finished yet.

        .. note::
            Some limitations exist for taking screenshots in 2D mode, since
            the underlying screenshot mechanism is the ArcGIS JS API's
            PrintTask object (https://bit.ly/2qRKGJG). If this function fails
            to create a screenshot, try a slightly different extent and try
            again. Non-published features with client-side graphics like 
            Spatially Enabled DataFrames might not display in a 2D mode 
            screenshot. 
        """
        if not self.ready:
            log.warn("Cannot take screenshot if widget is not visible in "\
                     "notebook: Please try again when widget is visible.");
            return False

        if not self.print_service_url:
            # If the print task URL hasn't been figured out yet, find it
            self._setup_screenshot_print_service_url()

        if output_in_cell:
            self._cell_output_screenshot_callback_resp = "" # Clear existing
            loading_html = HTML('<img style="float: left; padding-right: '\
                                '5px" src="' + _loading_icon_str + '"></img>'\
                                'Taking Screenshot: Please wait... </div>')
            self._cell_output_display_handler = display(loading_html,
                display_id = "cell-output-" + str(uuid4()))
        if file_path:
            self._file_output_screenshot_callback_resp = "" # Clear existing
            if not file_path.endswith(".png"):
                file_path = file_path + ".png"
            self._screenshot_file_output_path = file_path

        self._trigger_screenshot_with_args = \
            { '_' : str(uuid4()),
              'set_as_preview' : set_as_preview,
              'output_in_cell' : output_in_cell,
              'file_path' : bool(file_path) }

    # End screenshot specific section

    def _setup_gis_properties(self, gis):
        self.gis = gis
        if gis._portal.con._token:
            self._portal_token = str(gis._portal.con._token)
            self._auth_mode = "tokenBased"
        else:
            self._auth_mode = "anonymous"
        if gis._is_hosted_nb_home:
            #A GIS('home') conn needs to connect to public URL
            self._portal_url = str(gis._public_portal_url)
            self._portal_sharing_rest_url = self._portal_url + '/sharing/rest/'
        else:
            self._portal_url = str(gis._url)
            self._portal_sharing_rest_url = str(gis._con.baseurl)
        self._username = str(gis._username)

    def _setup_js_cdn(self):
        if _js_cdn_override_global != "":
            # If the user had previously set this global property, use it
            self._js_cdn_override = _js_cdn_override_global
        else:
            # Else, test default CDNs and portal CDNs
            default_cdn_unreachable = not self._is_reachable(_DEFAULT_JS_CDN)
            if default_cdn_unreachable:
                _portal_cdn = "{}/jsapi/jsapi4/".format(
                    getattr(self, "_portal_url", ""))
                _portal_cdn_unreachable = not self._is_reachable(_portal_cdn)
                if _portal_cdn_unreachable:
                    log.warn("Error: Cannot reach any JS API CDN, widget will"\
                             " not display".format(DEFAULT_JS_CDN))
                else:
                    self._js_cdn_override = _portal_cdn

    def _setup_default_basemap(self):
        """This method gets called once on startup, it populates the 'default'
        basemap field and the corresponding JSON without loading the rest
        of the `gallery_basemaps` property (which has a long load time)
        """
        if 'defaultBasemap' in self.gis.properties:
            self._gallery_basemaps['default'] = \
                self.gis.properties['defaultBasemap']
            self._basemap = 'default'
            # You need to re-write this dict to trigger the JS side change
            copy_gallery = dict(self._gallery_basemaps)
            self._gallery_basemaps = {}
            self._gallery_basemaps = copy_gallery

    def _is_reachable(self, url):
        import urllib.request
        def is_good_range(code):
            return (code >= 200) and (code < 300)
        try:
            conn = urllib.request.urlopen(url)
        except urllib.error.HTTPError as e:
            #  Return code error (e.g. 404, 501, ...)
            return is_good_range(e.code)
        except Exception as e:
            #  Not an HTTP-specific error (e.g. connection refused)
            #  as well as any malformed url passed in
            return False
        else:
            #  200 or other 'good' code
            return is_good_range(conn.code)

    def _check_if_webmap(self, item):
        from arcgis.gis import Item
        from arcgis.mapping import WebMap
        if isinstance(item, Item) and (item.type.lower() == 'web map'):
            item = WebMap(item)
        if isinstance(item, WebMap):
            if item.item.type.lower() == 'web map':
                self.webmap = item
                if hasattr(item.item, 'id'):
                    self._webmap = {"portalItem" : {
                        "id" : item.item.id } }
                    if hasattr(item.item, 'extent'):
                        self.extent = item.item.extent

    def _check_if_webscene(self, item):
        from arcgis.gis import Item
        if isinstance(item, Item):
            if item.type.lower() == 'web scene':
                self.webscene_item = item
                if hasattr(item, 'id'):
                    self._webscene = { "portalItem" : {
                        "id" : item.id } }

    @classmethod
    def set_js_cdn(cls, js_cdn):
        """Call this function before the creation of any MapView object, and
        each instantiated object will use the specified `js_cdn` parameter as
        the  ArcGIS API for JavaScript CDN URL instead of the default
        http://js.arcgis.com/4.X/. This functionality is necessary in
        disconnected  environments if the portal you are connecting to doesn't
        ship with the minimum necessary JavaScript API version. See the guide
        page on disconnected environments for more information."""
        global _js_cdn_override_global
        _js_cdn_override_global = js_cdn

    def add_layer(self, item, options=None):
        """
        Adds the specified layer or item to the map widget.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required object. You can specify Item objects, Layer objects such as
                               FeatureLayer, ImageryLayer, MapImageLayer etc., FeatureSet and
                               FeatureCollection objects.
        ------------------     --------------------------------------------------------------------
        options                Optional dict. Specify visualization options such as renderer info,
                               opacity, definition expressions. See example below
        ==================     ====================================================================

        .. code-block:: python

            # USAGE EXAMPLE: Add a feature layer with smart mapping renderer and
            # a definition expression to limit the features drawn.
            map1 = gis.map("Seattle, WA")
            map1.add_layer(wa_streets_feature_layer, {'renderer':'ClassedSizeRenderer',
                                                      'field_name':'DistMiles',
                                                      'opacity':0.75})

        """
        if options is None:
            options = {}
        self._add_layer_to_widget(item, options)

    def _add_layer_to_webmap(self, item, options):
        webmap_options = dict(options)
        webmap_options['layerId'] = self._get_hash(item)
        self.webmap.add_layer(item, webmap_options)

    def _add_layer_to_widget(self, item, options):
        from arcgis.features import FeatureSet, Feature, FeatureCollection
        from arcgis.raster import ImageryLayer
        from arcgis.gis import Layer
        from arcgis.gis import Item
        from arcgis._impl.common._mixins import PropertyMap

        if isinstance(item, Item):
            for layer in item.layers:
                self._add_layer_to_widget(layer, options)
        elif isinstance(item, Layer):
            self._add_layer_to_webmap(item, options)
            # TODO: Expand this to separate out Layer types on Python side
            # (i.e., do what was done for ImageryLayer for all major Layers)
            # 'No type' layer just means that we'll figure it out at JS time
            _lyr = _make_jsonable_dict(item._lyr_json)
            if ('type' in _lyr and \
                _lyr['type'] == 'MapImageLayer') and \
               ('TilesOnly' in item.properties.capabilities):
                # If it's a TilesOnly MapImageLayer, switch to TiledService
                _lyr['type'] = 'ArcGISTiledMapServiceLayer'
            if 'options' in _lyr:
                lyr_options = json.loads(_lyr["options"])
                lyr_options.update(options)
                _lyr["options"] = lyr_options
            else:
                _lyr["options"] = options
            _lyr["_hashFromPython"] = self._get_hash(item)
            self._add_notype_layer(item, _lyr)
        elif isinstance(item, FeatureSet):
            fc = FeatureCollection.from_featureset(item)
            self._add_layer_to_widget(fc, options)

        elif isinstance(item, dict):
            """Must have 'type' and 'url' in keys. Adds to notype_layers"""
            if ("type" not in item) or ("url" not in item):
                raise Exception("dict layers must have 'type' and 'url'")
            if "renderer" in options:
                item["renderer"] = options["renderer"]
            item["_hashFromPython"] = self._get_hash(item)
            self._add_notype_layer(item, item)
        elif _is_iterable(item):
            # If it's any iterable not previously checked, attempt to infer
            if 'layers' in item:
                for layer in item.layers:
                    self._add_layer_to_widget(layer, options)
            else:
                for item_ in item:
                    self._add_layer_to_widget(item_, options)
        else:
            raise RuntimeError("Cannot infer layer: will not be added to map")

    def _add_notype_layer(self, item, lyr_json):
        # Add the original item to the hashed layers
        self._add_to_hashed_layers(item)
        # but draw the json representation
        self._draw_these_notype_layers_on_widget_load += (lyr_json,)
        if self.ready:
            self._add_this_notype_layer = {}
            self._add_this_notype_layer = lyr_json

    def remove_layers(self, layers=None):
        """
        Removes the layers added to the map widget. You can get the list of layers added to the widget by querying the
        'layers' property.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        layers                 Optional list. Specify the list of layers to be removed from the map widget. You can get
                               the list of layers on the map by querying the 'layers' property.
                               If not specified, it removes all layers added to the widget.
        ==================     ====================================================================

        :return:
            True if layer is successfully removed. Else, False.
        """
        # Developer notes: infer all the layer types the user passed in.
        # Remove everything if the user didn't specify. Then, look up the hash
        # for each layer, remove it from the python side, trigger the removal
        # from the JS side

        output_bool = True
        if layers == None:
            layers = self.layers
        else:
            layers = self._infer_layers(layers)
        layer_hashes_to_remove = []
        for layer in layers:
            hash_ = self._get_hash(layer)
            if hash_ in self._hashed_layers:
                # Remove layer from py
                popped_layer = self._hashed_layers.pop(hash_, None)
                # stage remove from js
                layer_hashes_to_remove.append(hash_)
                # Remove layer from webmap
                for wm_layer in self.webmap.layers:
                    if wm_layer["id"] == hash_:
                        self.webmap.remove_layer(wm_layer)
            else:
                log.warn("Could not find layer {} in layers".format(layer))
                output_bool = False

        # Layer is removed from python side: trigger removal from JS side
        self._layers_to_remove = tuple('nonexistant_layer_id')
        self._layers_to_remove = tuple(layer_hashes_to_remove)

        return output_bool

    def _infer_layers(self, arg):
        """For a generic list of Layers, Items, FeatureSets, or an individual
        'layer', or anything, attempt to return a list of of 'layer' types
        that would exist in self.layers
        """
        from arcgis.features import FeatureSet, Feature, FeatureCollection
        from arcgis.raster import ImageryLayer
        from arcgis.gis import Layer
        from arcgis.gis import Item
        from arcgis._impl.common._mixins import PropertyMap

        output_layers = []
        if isinstance(arg, Layer):
            output_layers.append(arg)
        elif isinstance(arg, Item):
            for layer in arg.layers:
                output_layers.append(layer)
        elif isinstance(arg, FeatureSet):
            fc = FeatureCollection.from_featureset(arg)
            for layer in fc:
                output_layers.append(layer)
        elif isinstance(arg, dict):
            output_layers.append(arg)
        elif _is_iterable(arg):
            # If it's any iterable not previously checked, attempt to infer
            if 'layers' in arg:
                for layer in arg.layers:
                    output_layers.append(layer)
            else:
                for item_ in arg:
                    inferred_item = self._infer_layers(item_)[0]
                    output_layers.append(inferred_item)
        else:
            raise RuntimeError("Cannot infer layer {}: will not be removed"\
                               "".format(arg))
        return output_layers

    def _add_to_hashed_layers(self, item):
        hash_ = self._get_hash(item)
        self._hashed_layers[hash_] = item

    def _remove_from_notype_layers(self, layer):
        layers_to_test = list(self._notype_layers)
        for i in range(0, len(layers_to_test)):
            layer_to_test = layers_to_test[i]
            if layer_to_test['_hashFromPython'] == self._get_hash(layer): # ['_hashFromPython']:
                layers_to_test.pop(i)
        self._notype_layers = tuple(layers_to_test)

    def _get_hash(self, item):
        """Returns the str representation of the hash of any supported obj"""
        if isinstance(item, dict) and '_hashFromPython' in item:
            return item['_hashFromPython']
        if self._is_hashable(item):
            return str(hash(item))
        else:
            if isinstance(item, dict):
                return str(hash(frozenset(item)))
            else:
                raise Exception("Cannot hash item {}".format(item))

    def _is_hashable(self, item):
        try:
            hash(item)
            return True
        except Exception:
            return False

    def display_message(self, msg):
        """Displays a message on the upper-right corner of the map widget.
        You can only send one message at a time, multiple messages don’t show.
        """
        self._custom_msg = ""
        self._custom_msg = msg

    def save(self, item_properties, mode=None, thumbnail=None,
             metadata=None, owner=None, folder=None):
        """
        Save the map widget object into a new web map Item or a new web scene
        item in your GIS.

        .. note::
            If you started out with a fresh map widget object, use this method
            to save it as a the webmap/webscene item in your GIS.
            If you started with a map widget object from an existing
            webmap/webscene object, calling this method will create a new item
            with your changes. If you want to update the existing item with your
            changes, call the `update()` method instead.

        .. note::
            Saving as a WebScene item only works in a Jupyter environment:
            the map must be visually displayed in the notebook before
            calling this method.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_properties     Required dictionary. See table below for the keys and values.
        ---------------     --------------------------------------------------------------------
        mode                Optional string. Whether to save this map instance as a 2D WebMap,
                            or a 3D WebScene. Possible strings: "2D", "webmap", "3D", or
                            "webscene".
        ---------------     --------------------------------------------------------------------
        thumbnail           Optional string. Either a path or URL to a thumbnail image.
        ---------------     --------------------------------------------------------------------
        metadata            Optional string. Either a path or URL to the metadata.
        ---------------     --------------------------------------------------------------------
        owner               Optional string. User object corresponding to the desired owner of
                            this item. Defaults to the logged in user.
        ---------------     --------------------------------------------------------------------
        folder              Optional string. Name of the folder where placing item.
        ===============     ====================================================================

        *Key:Value Dictionary Options for Argument item_properties*

        =================  =====================================================================
        **Key**            **Value**
        -----------------  ---------------------------------------------------------------------
        typeKeywords       Optional string. Provide a lists all sub-types, see URL 1 below for
                           valid values.
        -----------------  ---------------------------------------------------------------------
        description        Optional string. Description of the item.
        -----------------  ---------------------------------------------------------------------
        title              Optional string. Name label of the item.
        -----------------  ---------------------------------------------------------------------
        tags               Optional string. Tags listed as comma-separated values, or a list of
                           strings. Used for searches on items.
        -----------------  ---------------------------------------------------------------------
        snippet            Optional string. Provide a short summary (limit to max 250
                           characters) of the what the item is.
        -----------------  ---------------------------------------------------------------------
        accessInformation  Optional string. Information on the source of the content.
        -----------------  ---------------------------------------------------------------------
        licenseInfo        Optional string.  Any license information or restrictions regarding
                           the content.
        -----------------  ---------------------------------------------------------------------
        culture            Optional string. Locale, country and language information.
        -----------------  ---------------------------------------------------------------------
        access             Optional string. Valid values are private, shared, org, or public.
        -----------------  ---------------------------------------------------------------------
        commentsEnabled    Optional boolean. Default is true, controls whether comments are
                           allowed (true) or not allowed (false).
        -----------------  ---------------------------------------------------------------------
        culture            Optional string. Language and country information.
        =================  =====================================================================

        URL 1: http://resources.arcgis.com/en/help/arcgis-rest-api/index.html# //02r3000000ms000000
        :return:
            Item object corresponding to the new web map Item created.

        .. code-block:: python

           USAGE EXAMPLE: Save map widget as a new web map item in GIS
           map1 = gis.map("Italy")
           map1.add_layer(Italy_streets_item)
           map1.basemap = 'dark-gray'
           italy_streets_map = map1.save({'title':'Italy streets',
                                        'snippet':'Arterial road network of Italy',
                                        'tags':'streets, network, roads'})
        """
        if mode == None:
            mode = self.mode
        if mode == "2D" or "webmap" in mode.lower():
            self.mode = "2D"
            return self._save_as_webmap(item_properties = item_properties,
                                        thumbnail = thumbnail,
                                        metadata = metadata,
                                        owner = owner,
                                        folder = folder)
        elif mode == "3D" or "webscene" in mode.lower():
            return self._save_as_webscene(item_properties,
                                          thumbnail = thumbnail,
                                          metadata = metadata,
                                          owner = owner,
                                          folder = folder)
        else:
            raise RuntimeError("'mode' arg must be one of '2D', 'webmap',"\
                               " '3D', or 'webscene'")

    def _save_as_webmap(self, item_properties, thumbnail=None,
                        metadata=None, owner=None, folder=None):
        from arcgis.mapping import WebMap
        self.mode = "2D"
        self._check_item_properties(item_properties)
        if ('basemap' in self._readonly_webmap_from_js) and \
           ('baseMapLayers' in self._readonly_webmap_from_js['basemap']):
            self.webmap._basemap['baseMapLayers'] = \
                self._readonly_webmap_from_js['basemap']['baseMapLayers']
        if self.extent:
            self.webmap._extent = self.extent
        self._update_webmap_layers_from_js()
        item = self.webmap.save(item_properties, thumbnail,
                                metadata, owner, folder)
        self.webmap.item = item
        return item

    def _update_webmap_layers_from_js(self):
        if self.ready:
            self._check_for_renderer_updates()
            self._check_for_drawn_layers()

    def _check_for_renderer_updates(self):
        js_layers = {layer["id"] : layer
                     for layer in self._readonly_webmap_from_js['layers']}
        for wm_layer in self.webmap.layers:
            wm_layer_id = wm_layer["id"]
            if wm_layer_id in js_layers:
                js_layer = js_layers[wm_layer_id] # js representation of wm layer
                if js_layer["renderer"]:
                    renderer = js_layer["renderer"]
                    id = js_layer["id"]
                    self._apply_renderer_to_webmap_layer_id(renderer, id)

    def _apply_renderer_to_webmap_layer_id(self, renderer, id):
        # TODO: find more elegant solution to solve this problem
        index = 0
        for layer in self.webmap._webmapdict["operationalLayers"]:
            if layer["id"] == id:
                wm_layer = self.webmap._webmapdict["operationalLayers"][index]
                if 'featureCollection' in wm_layer:
                    wm_layer = wm_layer['featureCollection']['layers'][0]
                wm_layer['layerDefinition']['drawingInfo'] = \
                    {'renderer' : renderer}
            else:
                index += 1

    def _check_for_drawn_layers(self):
        from arcgis.geometry import (Point, Polygon, Polyline,
                                     MultiPoint, Geometry)
        from arcgis.features import FeatureSet, Feature, FeatureCollection
        for layer in self._readonly_webmap_from_js['layers']:
            if ('graphics' in layer) and (len(layer['graphics']) > 0):
                for graphic in layer['graphics']:
                    # Infer what type of JS geometry it is
                    if 'shape' not in graphic and 'geometry' in graphic:
                        geom = Geometry(graphic['geometry'])
                    elif graphic['shape'] == 'polyline':
                        geom = Polyline(graphic['geometry'])
                    elif graphic['shape'] == 'polygon':
                        geom = Polygon(graphic['geometry'])
                    elif graphic['shape'] == 'multipoint':
                        geom = MultiPoint(graphic['geometry'])
                    elif graphic['shape']=='point':
                        geom = Point(graphic['geometry'])
                    elif 'geometry' in graphic:
                        geom = Geometry(graphic['geometry'])
                    else:
                        log.warn("Graphic unsupported, not adding to webmap."\
                                 " {}".format(graphic))
                        continue

                    # Create a python object from the JS geometry
                    feat = Feature(geom)
                    fset = FeatureSet([feat])

                    # Add to webmap
                    self.webmap.add_layer(fset,
                                          {'title' : 'Notes from ArcGIS API for Python'})


    def _save_as_webscene(self, item_properties, thumbnail=None,
                          metadata=None, owner=None, folder=None):
        self.mode = "3D"
        self._check_item_properties(item_properties)
        item_properties['type'] = 'Web Scene'
        if self.webscene_item:
            item_properties['text'] = self.webscene_item.get_data()
        else:
            item_properties['text'] = self._default_webscene_text_property

        self.webscene_item = self.gis.content.add(item_properties,
                                                  thumbnail = thumbnail,
                                                  metadata = metadata,
                                                  owner = owner,
                                                  folder = folder)
        self._trigger_webscene_save_to_this_portal_id = self.webscene_item.id
        return self.webscene_item

    def _check_item_properties(self, item_properties):
        if ('title' not in item_properties) or\
           ('snippet' not in item_properties) or\
           ('tags' not in item_properties):
            raise RuntimeError("title, snippet and tags are required in "\
                               "item_properties dictionary")

    def update(self, mode=None, item_properties=None,
               thumbnail=None, metadata=None):
        """
        Updates the web map item that was used to create the MapWidget object.
        In addition, you can update other item properties, thumbnail and
        metadata.

        .. note::
            If you started out a MapView object from an existing
            webmap/webscene item, use this method to update the
            webmap/webscene item in your with your changes.
            If you started out with a fresh MapView object (without a
            webmap/webscene item), calling this method will raise a
            RuntimeError exception. If you want to save the map widget into a
            new item, call the `save()` method instead.
            For item_properties, pass in arguments for only the properties you
            want to be updated. All other properties will be untouched.  For
            example, if you want to update only the item's description, then
            only provide the description argument in item_properties.

        .. note::
            Saving as a WebScene item only works in a Jupyter environment: the
            map must be visually displayed in the notebook before calling this
            method.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_properties     Optional dictionary. See table below for the keys and values.
        ---------------     --------------------------------------------------------------------
        mode                Optional string. Whether to save this map instance as a 2D WebMap,
                            or a 3D WebScene. Possible strings: "2D", "webmap", "3D", or
                            "webscene".
        ---------------     --------------------------------------------------------------------
        thumbnail           Optional string. Either a path or URL to a thumbnail image.
        ---------------     --------------------------------------------------------------------
        metadata            Optional string. Either a path or URL to the metadata.
        ===============     ====================================================================

        *Key:Value Dictionary Options for Argument item_properties*

        =================  =====================================================================
        **Key**            **Value**
        -----------------  ---------------------------------------------------------------------
        typeKeywords       Optional string. Provide a lists all sub-types, see URL 1 below for
                           valid values.
        -----------------  ---------------------------------------------------------------------
        description        Optional string. Description of the item.
        -----------------  ---------------------------------------------------------------------
        title              Optional string. Name label of the item.
        -----------------  ---------------------------------------------------------------------
        tags               Optional string. Tags listed as comma-separated values, or a list of
                           strings. Used for searches on items.
        -----------------  ---------------------------------------------------------------------
        snippet            Optional string. Provide a short summary (limit to max 250
                           characters) of the what the item is.
        -----------------  ---------------------------------------------------------------------
        accessInformation  Optional string. Information on the source of the content.
        -----------------  ---------------------------------------------------------------------
        licenseInfo        Optional string.  Any license information or restrictions regarding
                           the content.
        -----------------  ---------------------------------------------------------------------
        culture            Optional string. Locale, country and language information.
        -----------------  ---------------------------------------------------------------------
        access             Optional string. Valid values are private, shared, org, or public.
        -----------------  ---------------------------------------------------------------------
        commentsEnabled    Optional boolean. Default is true, controls whether comments are
                           allowed (true) or not allowed (false).
        =================  =====================================================================

        URL 1: http://resources.arcgis.com/en/help/arcgis-rest-api/index.html# //02r3000000ms000000
        :return:
           A boolean indicating success (True) or failure (False).

        .. code-block:: python

           USAGE EXAMPLE: Interactively add a new layer and change the basemap of an existing web map.
           italy_streets_item = gis.content.search("Italy streets", "Web Map")[0]
           map1 = MapView(item = italy_streets_item)
           map1.add_layer(Italy_streets2)
           map1.basemap = 'dark-gray-vector'
           map1.update(thumbnail = './new_webmap.png')

        """
        if mode == None:
            mode = self.mode
        if mode == "2D" or "webmap" in mode.lower():
            self.mode = "2D"
            return self._update_as_webmap(item_properties = item_properties,
                                          thumbnail = thumbnail,
                                          metadata = metadata)
        elif mode == "3D" or "webscene" in mode.lower():
            self.mode = "3D"
            return self._update_as_webscene(item_properties = item_properties,
                                            thumbnail = thumbnail,
                                            metadata = metadata)

    def _update_as_webmap(self, item_properties, thumbnail, metadata):
        from arcgis.mapping import WebMap
        if not self.webmap.item:
            raise RuntimeError("Webmap Item object missing. You should use "\
                               "`save()` to save a new web scene item")
        self.mode = "2D"
        self.webmap._basemap['baseMapLayers'] = \
            self._readonly_webmap_from_js['basemap']['baseMapLayers']
        self.webmap._extent = self.extent
        self._update_webmap_layers_from_js()
        return self.webmap.item.update(item_properties=item_properties,
                                       thumbnail=thumbnail,
                                           metadata=metadata)

    def _update_as_webscene(self, item_properties, thumbnail, metadata):
        if not self.webscene_item:
            raise RuntimeError("Webscene Item object missing. You should use "\
                               "`save()` to save a new web scene item")
        self.mode = "3D"
        result = self.webscene_item.update(item_properties=item_properties,
                                           thumbnail=thumbnail,
                                           metadata=metadata)
        self._trigger_webscene_save_to_this_portal_id = ""
        self._trigger_webscene_save_to_this_portal_id = self.webscene_item.id
        return result

    def export_to_html(self, path_to_file, title="Exported ArcGIS Map Widget",
                       credentials_prompt=False):
        """
        Takes the current state of the map widget, and exports it to a
        standalone HTML file that can be viewed in any web browser.

        By default, only publically viewable layers will be visible in any
        exported html map. Specify `credentials_prompt=True` to have a user
        be prompted for their credentials when opening the HTML page to view
        private content.

        .. warning::
            Follow best security practices when sharing any HTML page that
            prompts a user for a password.

        .. note::
            You cannot successfully authenticate if you open the HTML page in a
            browser locally like file://path/to/file.html. The credentials
            prompt will only properly function if served over a HTTP/HTTPS
            server.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        path_to_file           Required string. The path to save the HTML file on disk.
        ------------------     --------------------------------------------------------------------
        title                  Optional string. The HTML title tag used for the HTML file.
        ------------------     --------------------------------------------------------------------
        credentials_prompt     Optional boolean, default `False`. If set to `True`, will display a
                               credentials prompt on HTML page load for users to authenticate and
                               view private content.
        ==================     ====================================================================
        """
        # Make sure height is to 100% so exported page loads in fullscreen
        prev_height = self.layout.height
        self.layout.height = "100%"

        # Set the _auth_mode
        prev_auth_mode = self._auth_mode
        if credentials_prompt:
            self._auth_mode = "prompt"
        else:
            self._auth_mode = "anonymous"

        # 'extent' is a property with separate readonly/writeonly fields
        # The below statement makes the readonly match the writeonly
        # This makes sure the current view is exactly what is saved in the
        # model for the exported HTML
        self.extent = self.extent

        # Write the HTML file
        embed_minimal_html(path_to_file, views=[self], title=title)

        # Restore the previous model attributes that were changed
        self._auth_mode = prev_auth_mode
        self.layout.height = prev_height
        return True

    def draw(self, shape, popup=None,
             symbol=None, attributes=None):
        """
        Draws a shape on the map widget. You can draw anything from known geometries, coordinate pairs, FeatureSet
        objects.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        shape                  Required object.
                               Known geometries:
                               Shape is one of ["circle", "ellipse", "polygon", "polyline",
                               "multipoint", "point", "rectangle", "triangle"].
                               Coordinate pair: specify shape as a list of [lat, long]. Eg: [34, -81]
                               FeatureSet: shape can be a FeatureSet object.
                               Dict object representing a geometry.
        ------------------     --------------------------------------------------------------------
        popup                  Optional dict. Dict containing "title" and "content" as keys that will be displayed
                               when the shape is clicked. In case of a FeatureSet, "title" and "content" are names of
                               attributes of the features in the FeatureSet instead of actual string values for title
                               and content.
        ------------------     --------------------------------------------------------------------
        symbol                 Optional dict. symbol is specified in json format as described at
                               http://resources.arcgis.com/en/help/arcgis-rest-api/index.html# //02r3000000n5000000. A
                               default symbol is used if one is not specified.
                               Tip: a helper utility to get the symbol format for several predefined symbols is
                               available at http://esri.github.io/arcgis-python-api/tools/symbol.html
        ------------------     --------------------------------------------------------------------
        attributes             Optional dict. Specify a dict containing name value pairs of fields and field values
                               associated with the graphic.
        ==================     ====================================================================
        """
        from arcgis.features import FeatureSet, Feature, FeatureCollection
        from arcgis.raster import ImageryLayer
        from arcgis.gis import Layer
        from arcgis.gis import Item
        from arcgis._impl.common._mixins import PropertyMap

        title = attributes['title'] if attributes and \
            'title' in attributes else "Notebook sketch layer"

        if isinstance(shape, list) and len(shape) == 2:  #  [lat, long] pair
            shape = {'x': shape[1], 'y': shape[0],
                     "spatialReference": {"wkid": 4326}, 'type': 'point'}

        elif isinstance(shape, tuple):  #  (lat, long) pair
            shape = {'x': shape[1], 'y': shape[0],
                     "spatialReference": {"wkid": 4326}, 'type': 'point'}

        elif isinstance(shape, dict) and 'location' in shape: #  geocode loc.
            shape = {'x': shape['location']['x'], 'y': shape['location']['y'],
                     "spatialReference": {"wkid": 4326}, 'type': 'point'}

        if isinstance(shape, FeatureSet) or isinstance(shape, dict):
            fset = None
            if isinstance(shape, FeatureSet):
                fset = shape
                self._draw_featureset(fset, popup, symbol)
            elif isinstance(shape, dict):
                if 'type' not in shape:
                    if 'rings' in shape:
                        shape['type'] = 'polygon'
                    elif 'paths' in shape:
                        shape['type'] = 'polyline'
                    elif 'points' in shape:
                        shape['type'] = 'multipoint'
                    else:
                        shape['type'] = 'point'
                graphic = {
                    "geometry": shape,
                    "popupTemplate": popup,
                    "symbol": symbol,
                    "attributes": attributes
                }
                self._add_graphic(graphic)
                f = Feature(shape)
                fset = FeatureSet([f],
                                  geometry_type='esriGeometryPoint',
                                  spatial_reference={'wkid':4326})

            # Now that the `fset` is set, add to webmap
            if popup:
                webmap_popup = {'title':"{" +popup['title']+"}",
                                'description':"{" + popup['content'] +"}"}
            else:
                webmap_popup = None
            wm_options = {'popup':webmap_popup, 'symbol':symbol,
                          'attributes':attributes, 'title':title,
                          'extent':self.extent}
            self.webmap.add_layer(fset, wm_options)

        else: # User passed in a string for interactive draw mode
            self._trigger_interactive_draw_mode_for = ""
            self._trigger_interactive_draw_mode_for = shape

    def _draw_featureset(self, fset, popup, symbol):
        # FeatureSet needs special case
        graphics = []
        for feature in fset.features:
            graphic = self._get_graphic_from_feature(feature,
                                                         popup, symbol)
            graphics.append(graphic)
            if self.ready:
                self._add_this_graphic = {}
                self._add_this_graphic = graphic
        self._draw_these_graphics_on_widget_load += tuple(graphics)

    def _get_graphic_from_feature(self, feature, popup, symbol):
        if popup:
            popup_dict = \
                {'title' : feature.attributes[popup['title']],
                 'content' : feature.attributes[popup['content']]}
        else:
            popup_dict = None
        geometry = feature.geometry
        geometry['type'] = feature.geometry_type.lower()
        graphic = {
            "geometry": feature.geometry,
            "popupTemplate": popup_dict,
            "symbol": symbol,
            "attributes": feature.attributes
        }
        return _make_jsonable_dict(graphic)

    def _add_graphic(self, graphic):
        if self.ready:
            self._add_this_graphic = {}
            self._add_this_graphic = _make_jsonable_dict(graphic)
        else:
            self._draw_these_graphics_on_widget_load += \
                (_make_jsonable_dict(graphic),)

    def clear_graphics(self):
        """
        Clear the graphics drawn on the map widget. Graphics are shapes drawn
        using the 'draw()' method.
        """
        self._layers_to_remove = ("nonexistant_layer_id",)
        # All graphics are saved to a layer with the below id
        self._layers_to_remove = ("graphicsLayerId31195",)

    def on_draw_end(self, callback, remove=False):
        """Register a callback to execute when something is drawn.
        The callback will be called with two arguments:
        the clicked widget instance, and the geometry drawn

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        remove                 Optional boolean. Set to true to remove the callback from the list
                               of callbacks.
        ==================     ====================================================================

        """
        self._draw_end_handlers.register_callback(callback, remove=remove)

    def on_click(self, callback, remove=False):
        """Register a callback to execute when the map is clicked.
        The callback will be called with one argument,
        the clicked widget instance.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        remove                 Optional boolean. Set to true to remove the callback from the list
                               of callbacks.
        ==================     ====================================================================

        """
        self._click_handlers.register_callback(callback, remove=remove)

    def toggle_window_view(self, title="ArcGIS Map", tab_mode=None):
        """In a JupyterLab environment, calling this function will separate
        the drawn map widget to a new window next to the open notebook,
        allowing you to move the widget it, split it, put it in a new tab, etc.
        If the widget is already seperated in a new window, calling this
        function will restore the widget to the notebook where it originated
        from. See the JupyterLab guide pages for more information.

        .. raw:: html

            <p>Note that this functionality can also be achieved by pressing
            the <img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAA
            gCAYAAABzenr0AAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAAALEwEAmpwYAA
            AAB3RJTUUH4gcUAAIIxu7mQQAAAB1pVFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoI
            EdJTVBkLmUHAAABI0lEQVRYw+1Wy3GDMBTch3JLBanA+TQR0gDz4JB0Y1xISkBAA4Eq
            HOeYU6qAzYVkPAyWYSLii/Ymzb63q89qBAQEXBgyNamquYhs5zSw1v72SNOUp3gkd2V
            Z5uP5K1dzkrslqxnzRWQD4MVV4zQw5XguP8uyDckGwBeAm1M10Rrnqqq3fd+3JEVEYh
            c3WkMcQDMM46IoPv7NQJIkdyLS/oiXZXk4VxP5FDfGNEvEz17Cuciy7J7kGwB0XRfXd
            X0Yp+NoZ/waGMQbABzE35ekKfIkDhF5mhJf7Q6o6sOReFwUxd7BzdM0ffR5BALglaQA
            iK21eydZZEsSAFpfBkjy2Rhz7Vr5qimoqurTR4JWeYqDgWDAWwpUNb+ogbn/woCAgL/
            gG9knfWwKBGcvAAAAAElFTkSuQmCC"> icon in the widget UI.</p>

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        title                  What text will display as the widget tab. Default: "ArcGIS Map".
        ------------------     --------------------------------------------------------------------
        tab_mode               The 'tab mode' that this window will open in. Will use this
                               MapView instance's `tab_mode` property if not specified. Possible
                               values: "auto", "split-top", "split-left", "split-right",
                               "split-bottom", "tab-before", "tab-after"
        ==================     ====================================================================
        """

        args = {"title": title}
        if tab_mode:
            args["tab_mode"] = tab_mode
        if not self.ready:
            self._ipython_display_()
            # TODO: Find more elegant solution for this time.sleep hack
            time.sleep(1)
        self._trigger_new_jlab_window_with_args = {}
        self._trigger_new_jlab_window_with_args = args

    def _handle_map_msg(self, _, content, buffers):
        """Handle a msg from the front-end.
        Parameters
        ----------
        content: dict
            Content of the msg."""

        if content.get('event', '') == 'mouseclick':
            self._click_handlers(self, content.get('message', None))
        if content.get('event', '') == 'draw-end':
            self._draw_end_handlers(self, content.get('message', None))

    # Start section of no longer supported areas
    def _raise_time_extent_exception(self):
        raise Exception("Time extent functionality not supported in v1.5")

    @property
    def end_time(self):
        """This feature is not supported in v1.5."""
        self._raise_time_extent_exception()

    @end_time.setter
    def end_time(self, value):
        self._raise_time_extent_exception()

    def set_time_extent(self, start_time, end_time):
        """This feature is not supported in v1.5."""
        self._raise_time_extent_exception()

    @property
    def start_time(self):
        """This feature is not supported in v1.5."""
        self._raise_time_extent_exception()

    @start_time.setter
    def start_time(self, value):
        self._raise_time_extent_exception()

    # end section of no longer supported areas

